"""License action definitions.

Version Added:
    8.0
"""

from __future__ import annotations

from typing import Mapping, Protocol, TYPE_CHECKING

from djblets.util.typing import JSONValue
from typing_extensions import TypedDict

if TYPE_CHECKING:
    from django.http import HttpRequest
    from djblets.util.typing import SerializableJSONDict, StrOrPromise
    from typelets.json import JSONDict
    from typing_extensions import NotRequired, TypeAlias

    from reviewboard.licensing.license import LicenseInfo


#: Data passed to an action handler.
#:
#: This is a mapping of string keys to JSON values.
#:
#: Version Added:
#:     8.0
LicenseActionData: TypeAlias = Mapping[str, JSONValue]


class LicenseAction(TypedDict):
    """An action that can be performed on a license.

    These will be displayed in the UI when displaying license information.

    Version Added:
        8.0
    """

    #: The provider-unique ID of the action.
    action_id: str

    #: The localized label for the action.
    label: StrOrPromise

    #: Arguments to pass in a call to the License Provider's action handler.
    call_args: NotRequired[JSONDict]

    #: Extra data to include in the action for client-side handling.
    extra_data: NotRequired[JSONDict]

    #: Whether this is a primary button.
    primary: NotRequired[bool]

    #: The URL to navigate to when clicking the action.
    #:
    #: If omitted or ``None``, it's up to the license implementation to
    #: provide JavaScript management for the action.
    url: NotRequired[str | None]


class LicenseActionHandler(Protocol):
    """Protocol for a method used to handle a license action request.

    This is the signature required for any :samp:`handle_{action}_action`
    methods on a license provider.

    Version Added:
        8.0
    """

    def __call__(
        self,
        *,
        license_info: LicenseInfo,
        action_data: LicenseActionData,
        request: HttpRequest | None,
    ) -> SerializableJSONDict:
        ...



"""Unit tests for repository operations."""

from __future__ import annotations

import os

import kgb
from django_assert_queries.testing import assert_queries
from django.contrib.auth.models import AnonymousUser
from django.core.exceptions import ImproperlyConfigured, ValidationError
from django.db.models import Q
from djblets.testing.decorators import add_fixtures

from reviewboard.hostingsvcs.errors import MissingHostingServiceError
from reviewboard.hostingsvcs.github import GitHub
from reviewboard.hostingsvcs.models import HostingServiceAccount
from reviewboard.scmtools.core import FileLookupContext
from reviewboard.scmtools.git import GitTool
from reviewboard.scmtools.models import Repository, Tool
from reviewboard.scmtools.signals import (checked_file_exists,
                                          checking_file_exists,
                                          fetched_file, fetching_file)
from reviewboard.testing.testcase import TestCase


class RepositoryTests(kgb.SpyAgency, TestCase):
    """Unit tests for Repository operations."""

    fixtures = ['test_scmtools']

    def setUp(self):
        super(RepositoryTests, self).setUp()

        self.local_repo_path = os.path.join(os.path.dirname(__file__),
                                            '..', 'testdata', 'git_repo')
        self.repository = self.create_repository(
            name='Git test repo',
            path=self.local_repo_path,
            tool_name='Git')

    def test_archive(self):
        """Testing Repository.archive"""
        repository1 = self.repository

        repository1.archive()
        self.assertTrue(repository1.name.startswith('ar:Git test repo:'))
        self.assertTrue(repository1.archived)
        self.assertFalse(repository1.public)
        self.assertIsNotNone(repository1.archived_timestamp)

        repository2 = Repository.objects.get(pk=repository1.pk)
        self.assertEqual(repository2.name,
                         repository1.name)
        self.assertEqual(repository2.archived,
                         repository1.archived)
        self.assertEqual(repository2.public,
                         repository1.public)
        self.assertEqual(repository2.archived_timestamp,
                         repository1.archived_timestamp)

    def test_archive_no_save(self):
        """Testing Repository.archive with save=False"""
        repository1 = self.repository

        repository1.archive(save=False)
        self.assertTrue(repository1.name.startswith('ar:Git test repo:'))
        self.assertTrue(repository1.archived)
        self.assertFalse(repository1.public)
        self.assertIsNotNone(repository1.archived_timestamp)

        repository2 = Repository.objects.get(pk=repository1.pk)
        self.assertNotEqual(repository2.name,
                            repository1.name)
        self.assertNotEqual(repository2.archived,
                            repository1.archived)
        self.assertNotEqual(repository2.public,
                            repository1.public)
        self.assertNotEqual(repository2.archived_timestamp,
                            repository1.archived_timestamp)

    def test_clean_without_conflict(self):
        """Testing Repository.clean without name/path conflicts"""
        with self.assertNumQueries(1):
            self.repository.clean()

    def test_clean_with_name_conflict(self):
        """Testing Repository.clean with name conflict"""
        repository = Repository(name=self.repository.name,
                                path='path/to/repo.git',
                                tool=self.repository.tool,
                                scmtool_id=self.repository.scmtool_id)

        with self.assertRaises(ValidationError) as ctx:
            with self.assertNumQueries(1):
                repository.clean()

        self.assertEqual(ctx.exception.message_dict, {
            'name': ['A repository with this name already exists'],
        })

    def test_clean_with_path_conflict(self):
        """Testing Repository.clean with path conflict"""
        repository = Repository(name='New test repo',
                                path=self.repository.path,
                                tool=self.repository.tool,
                                scmtool_id=self.repository.scmtool_id)

        with self.assertRaises(ValidationError) as ctx:
            with self.assertNumQueries(1):
                repository.clean()

        self.assertEqual(ctx.exception.message_dict, {
            'path': ['A repository with this path already exists'],
        })

    def test_clean_with_name_and_path_conflict(self):
        """Testing Repository.clean with name and path conflict"""
        repository = Repository(name=self.repository.name,
                                path=self.repository.path,
                                tool=self.repository.tool,
                                scmtool_id=self.repository.scmtool_id)

        with self.assertRaises(ValidationError) as ctx:
            with self.assertNumQueries(1):
                repository.clean()

        self.assertEqual(ctx.exception.message_dict, {
            'name': ['A repository with this name already exists'],
            'path': ['A repository with this path already exists'],
        })

    def test_clean_with_path_conflict_with_archived(self):
        """Testing Repository.clean with archived repositories ignored for
        path conflict
        """
        orig_repository = self.repository
        orig_repository.archive()

        repository = Repository(name='New test repo',
                                path=orig_repository.path,
                                tool=orig_repository.tool,
                                scmtool_id=self.repository.scmtool_id)

        with self.assertNumQueries(1):
            repository.clean()

    def test_get_file_with_context(self):
        """Testing Repository.get_file with context="""
        repository = self.repository
        scmtool_cls = repository.scmtool_class

        self.spy_on(scmtool_cls.get_file,
                    owner=scmtool_cls,
                    op=kgb.SpyOpReturn(b'data'))

        context = FileLookupContext(base_commit_id='def456')

        repository.get_file(path='readme',
                            revision='abc123',
                            context=context)

        self.assertSpyCalledWith(
            scmtool_cls.get_file,
            'readme',
            revision='abc123',
            base_commit_id='def456',
            context=context)

    def test_get_file_caching(self):
        """Testing Repository.get_file caches result"""
        path = 'readme'
        revision = 'abc123'
        base_commit_id = 'def456'

        repository = self.repository
        scmtool_cls = repository.scmtool_class

        self.spy_on(scmtool_cls.get_file,
                    owner=scmtool_cls,
                    op=kgb.SpyOpReturn(b'file data'))

        # Two requests to the same path/revision should result in one only
        # call.
        data1 = repository.get_file(path=path,
                                    revision=revision)
        data2 = repository.get_file(path=path,
                                    revision=revision,
                                    context=FileLookupContext())

        self.assertIsInstance(data1, bytes)
        self.assertIsInstance(data2, bytes)
        self.assertEqual(data1, b'file data')
        self.assertEqual(data1, data2)
        self.assertSpyCallCount(scmtool_cls.get_file, 1)
        self.assertSpyLastCalledWith(scmtool_cls.get_file,
                                     path,
                                     revision=revision,
                                     base_commit_id=None)

        # A base commit ID should result in a new call.
        data3 = repository.get_file(path=path,
                                    revision=revision,
                                    base_commit_id=base_commit_id)

        self.assertEqual(data3, data1)
        self.assertSpyCallCount(scmtool_cls.get_file, 2)
        self.assertSpyLastCalledWith(scmtool_cls.get_file,
                                     path,
                                     revision=revision,
                                     base_commit_id=base_commit_id)

        # Another fetch with the same base_commit_id will use the cached
        # version, even if specified in a FileLookupContext.
        context = FileLookupContext(base_commit_id=base_commit_id)
        data4 = repository.get_file(path=path,
                                    revision=revision,
                                    context=context)

        self.assertEqual(data4, data1)
        self.assertSpyCallCount(scmtool_cls.get_file, 2)

    def test_get_file_signals(self):
        """Testing Repository.get_file emits signals"""
        def on_fetching_file(**kwargs):
            pass

        def on_fetched_file(**kwargs):
            pass

        repository = self.repository

        fetching_file.connect(on_fetching_file, sender=repository)
        fetched_file.connect(on_fetched_file, sender=repository)

        self.spy_on(on_fetching_file)
        self.spy_on(on_fetched_file)

        path = 'readme'
        revision = 'e965047'
        base_commit_id = 'def456'

        request = self.create_http_request()
        context = FileLookupContext(request=request,
                                    base_commit_id=base_commit_id)

        repository.get_file(path=path,
                            revision=revision,
                            context=context)

        self.assertSpyCalledWith(
            on_fetching_file,
            sender=repository,
            path=path,
            revision=revision,
            base_commit_id=base_commit_id,
            request=request,
            context=context)

        self.assertSpyCalledWith(
            on_fetched_file,
            sender=repository,
            path=path,
            revision=revision,
            base_commit_id=base_commit_id,
            request=request,
            context=context,
            data=b'Hello\n')

    def test_get_file_exists_caching_when_exists(self):
        """Testing Repository.get_file_exists caches result when exists"""
        path = 'readme'
        revision = 'e965047'
        base_commit_id = 'def456'

        repository = self.repository
        scmtool_cls = repository.scmtool_class

        self.spy_on(scmtool_cls.file_exists,
                    owner=scmtool_cls,
                    op=kgb.SpyOpReturn(True))

        # Two requests to the same path/revision should result in one only
        # call.
        self.assertTrue(repository.get_file_exists(
            path=path,
            revision=revision))
        self.assertTrue(repository.get_file_exists(
            path=path,
            revision=revision,
            context=FileLookupContext()))

        self.assertSpyCallCount(scmtool_cls.file_exists, 1)
        self.assertSpyLastCalledWith(scmtool_cls.file_exists,
                                     path,
                                     revision=revision,
                                     base_commit_id=None)

        # A base commit ID should result in a new call.
        self.assertTrue(repository.get_file_exists(
            path=path,
            revision=revision,
            base_commit_id=base_commit_id))

        self.assertSpyCallCount(scmtool_cls.file_exists, 2)
        self.assertSpyLastCalledWith(scmtool_cls.file_exists,
                                     path,
                                     revision=revision,
                                     base_commit_id=base_commit_id)

        # Another check with the same base_commit_id will use the cached
        # version, even if specified in a FileLookupContext.
        context = FileLookupContext(base_commit_id=base_commit_id)
        self.assertTrue(repository.get_file_exists(
            path=path,
            revision=revision,
            context=context))

        self.assertSpyCallCount(scmtool_cls.file_exists, 2)

    def test_get_file_exists_caching_when_not_exists(self):
        """Testing Repository.get_file_exists doesn't cache result when the
        file does not exist
        """
        path = 'readme'
        revision = '12345'
        base_commit_id = 'def456'

        repository = self.repository
        scmtool_cls = repository.scmtool_class

        self.spy_on(scmtool_cls.file_exists,
                    owner=scmtool_cls,
                    op=kgb.SpyOpReturn(False))

        context = FileLookupContext(base_commit_id=base_commit_id)

        self.assertFalse(repository.get_file_exists(
            path=path,
            revision=revision))
        self.assertFalse(repository.get_file_exists(
            path=path,
            revision=revision,
            context=FileLookupContext()))
        self.assertFalse(repository.get_file_exists(
            path=path,
            revision=revision,
            base_commit_id=base_commit_id))
        self.assertFalse(repository.get_file_exists(
            path=path,
            revision=revision,
            context=context))

        self.assertSpyCallCount(scmtool_cls.file_exists, 4)
        self.assertSpyCalledWith(
            scmtool_cls.file_exists.calls[0],
            path,
            revision=revision,
            base_commit_id=None)
        self.assertSpyCalledWith(
            scmtool_cls.file_exists.calls[1],
            path,
            revision=revision,
            base_commit_id=None)
        self.assertSpyCalledWith(
            scmtool_cls.file_exists.calls[2],
            path,
            revision=revision,
            base_commit_id=base_commit_id)
        self.assertSpyCalledWith(
            scmtool_cls.file_exists.calls[3],
            path,
            revision=revision,
            base_commit_id=base_commit_id,
            context=context)

    def test_get_file_exists_caching_with_fetched_file(self):
        """Testing Repository.get_file_exists uses get_file's cached result"""
        path = 'readme'
        revision = 'abc123'
        base_commit_id = 'def456'

        repository = self.repository
        scmtool_cls = repository.scmtool_class

        self.spy_on(scmtool_cls.get_file,
                    owner=scmtool_cls,
                    op=kgb.SpyOpReturn(b'file data'))
        self.spy_on(scmtool_cls.file_exists,
                    owner=scmtool_cls,
                    op=kgb.SpyOpReturn(True))

        # These requests to the same path/revision should result in one only
        # call.
        repository.get_file(path=path,
                            revision=revision)
        self.assertTrue(repository.get_file_exists(
            path=path,
            revision=revision))
        self.assertTrue(repository.get_file_exists(
            path=path,
            revision=revision,
            context=FileLookupContext()))

        self.assertSpyCallCount(scmtool_cls.get_file, 1)
        self.assertSpyNotCalled(scmtool_cls.file_exists)

        # A base commit ID should result in a new call, which should then
        # persist for file checks.
        repository.get_file(path=path,
                            revision=revision,
                            base_commit_id=base_commit_id)
        self.assertTrue(repository.get_file_exists(
            path=path,
            revision=revision,
            base_commit_id=base_commit_id))
        self.assertTrue(repository.get_file_exists(
            path=path,
            revision=revision,
            context=FileLookupContext(base_commit_id=base_commit_id)))

        self.assertSpyCallCount(scmtool_cls.get_file, 2)
        self.assertSpyNotCalled(scmtool_cls.file_exists)

    def test_get_file_exists_signals(self):
        """Testing Repository.get_file_exists emits signals"""
        def on_checking(**kwargs):
            pass

        def on_checked(**kwargs):
            pass

        repository = self.repository

        checking_file_exists.connect(on_checking, sender=repository)
        checked_file_exists.connect(on_checked, sender=repository)

        self.spy_on(on_checking)
        self.spy_on(on_checked)

        path = 'readme'
        revision = 'e965047'
        base_commit_id = 'def456'

        request = self.create_http_request()
        context = FileLookupContext(request=request,
                                    base_commit_id=base_commit_id)

        repository.get_file_exists(path=path,
                                   revision=revision,
                                   context=context)

        self.assertSpyCalledWith(
            on_checking,
            sender=repository,
            path=path,
            revision=revision,
            base_commit_id=base_commit_id,
            request=request,
            context=context)

        self.assertSpyCalledWith(
            on_checked,
            sender=repository,
            path=path,
            revision=revision,
            base_commit_id=base_commit_id,
            request=request,
            context=context)

    def test_hosting_service(self):
        """Testing Repository.hosting_service with a valid hosting service"""
        account = HostingServiceAccount.objects.create(
            service_name=GitHub.hosting_service_id,
            username='foo')
        repository = self.create_repository(hosting_account=account)

        with self.assertNumQueries(0):
            self.assertIsInstance(repository.hosting_service, GitHub)

    def test_hosting_service_cannot_load(self):
        """Testing Repository.hosting_service with a hosting service
        that cannot be loaded
        """
        account = HostingServiceAccount.objects.create(
            service_name='blah',
            username='foo')
        repository = self.create_repository(hosting_account=account)
        message = ('The repository "Test Repo" cannot load the hosting '
                   'service "blah". An administrator should ensure all '
                   'necessary packages and extensions are installed.')

        with self.assertNumQueries(0):
            with self.assertRaisesMessage(MissingHostingServiceError, message):
                repository.hosting_service

    def test_repository_name_with_255_characters(self):
        """Testing Repository.name with 255 characters"""
        repository = self.create_repository(name='t' * 255)

        self.assertEqual(len(repository.name), 255)

    def test_is_accessible_by_with_public(self):
        """Testing Repository.is_accessible_by with public repository"""
        user = self.create_user()
        repository = self.create_repository()

        self.assertTrue(repository.is_accessible_by(user))
        self.assertTrue(repository.is_accessible_by(AnonymousUser()))

    def test_is_accessible_by_with_public_and_hidden(self):
        """Testing Repository.is_accessible_by with public hidden repository"""
        user = self.create_user()
        repository = self.create_repository(visible=False)

        self.assertTrue(repository.is_accessible_by(user))
        self.assertTrue(repository.is_accessible_by(AnonymousUser()))

    def test_is_accessible_by_with_private_and_not_member(self):
        """Testing Repository.is_accessible_by with private repository and
        user not a member
        """
        user = self.create_user()
        repository = self.create_repository(public=False)

        self.assertFalse(repository.is_accessible_by(user))
        self.assertFalse(repository.is_accessible_by(AnonymousUser()))

    def test_is_accessible_by_with_private_and_member(self):
        """Testing Repository.is_accessible_by with private repository and
        user is a member
        """
        user = self.create_user()
        repository = self.create_repository(public=False)
        repository.users.add(user)

        self.assertTrue(repository.is_accessible_by(user))

    def test_is_accessible_by_with_private_and_member_by_group(self):
        """Testing Repository.is_accessible_by with private repository and
        user is a member by group
        """
        user = self.create_user()

        group = self.create_review_group(invite_only=True)
        group.users.add(user)

        repository = self.create_repository(public=False)
        repository.review_groups.add(group)

        self.assertTrue(repository.is_accessible_by(user))

    def test_is_accessible_by_with_private_and_superuser(self):
        """Testing Repository.is_accessible_by with private repository and
        user is a superuser
        """
        user = self.create_user(is_superuser=True)
        repository = self.create_repository(public=False)

        self.assertTrue(repository.is_accessible_by(user))

    def test_is_accessible_by_with_private_hidden_not_member(self):
        """Testing Repository.is_accessible_by with private hidden
        repository and user not a member
        """
        user = self.create_user()
        repository = self.create_repository(public=False,
                                            visible=False)

        self.assertFalse(repository.is_accessible_by(user))

    def test_is_accessible_by_with_private_hidden_and_member(self):
        """Testing Repository.is_accessible_by with private hidden
        repository and user is a member
        """
        user = self.create_user()

        repository = self.create_repository(public=False,
                                            visible=False)
        repository.users.add(user)

        self.assertTrue(repository.is_accessible_by(user))

    def test_is_accessible_by_with_private_hidden_and_member_by_group(self):
        """Testing Repository.is_accessible_by with private hidden
        repository and user is a member
        """
        user = self.create_user()

        group = self.create_review_group(invite_only=True)
        group.users.add(user)

        repository = self.create_repository(public=False,
                                            visible=False)
        repository.review_groups.add(group)

        self.assertTrue(repository.is_accessible_by(user))

    def test_is_accessible_by_with_private_hidden_and_superuser(self):
        """Testing Repository.is_accessible_by with private hidden
        repository and superuser
        """
        user = self.create_user(is_superuser=True)
        repository = self.create_repository(public=False,
                                            visible=False)

        self.assertTrue(repository.is_accessible_by(user))

    @add_fixtures(['test_users', 'test_site'])
    def test_is_accessible_by_with_local_site_accessible(self):
        """Testing Repository.is_accessible_by with Local Site accessible by
        user
        """
        user = self.create_user()

        repository = self.create_repository(with_local_site=True)
        repository.local_site.users.add(user)

        self.assertTrue(repository.is_accessible_by(user))

    @add_fixtures(['test_users', 'test_site'])
    def test_is_accessible_by_with_local_site_not_accessible(self):
        """Testing Repository.is_accessible_by with Local Site not accessible
        by user
        """
        user = self.create_user()
        repository = self.create_repository(with_local_site=True)

        self.assertFalse(repository.is_accessible_by(user))
        self.assertFalse(repository.is_accessible_by(AnonymousUser()))

    def test_scmtool_class(self):
        """Testing Repository.scmtool_class"""
        with self.assertNumQueries(0):
            repository = Repository(scmtool_id='git')

            self.assertIs(repository.scmtool_class, GitTool)

    def test_scmtool_class_with_unset_scmtool_id(self):
        """Testing Repository.scmtool_class with unset scmtool_id and with
        legacy tool_id
        """
        tool = Tool.objects.get(name='Git')

        # 1 query:
        #
        # 1. Insert new repository.
        queries = [
            {
                'type': 'INSERT',
                'model': Repository,
            },
        ]

        with assert_queries(queries):
            repository = Repository.objects.create(name='test-repo',
                                                   tool=tool)

            self.assertIs(repository.scmtool_class, GitTool)
            self.assertEqual(repository.scmtool_id, 'git')

        # Make sure that persists.
        repository.refresh_from_db()

        with self.assertNumQueries(0):
            self.assertIs(repository.scmtool_class, GitTool)
            self.assertEqual(repository.scmtool_id, 'git')

    def test_scmtool_class_with_unset_scmtool_id_deferred(self):
        """Testing Repository.scmtool_class with unset scmtool_id and with
        legacy tool_id and deferred fields
        """
        tool = Tool.objects.get(name='Git')
        repository = Repository.objects.create(name='test-repo',
                                               tool=tool)
        repository.scmtool_id = None
        repository.save(update_fields=('scmtool_id',))

        # Re-fetch with deferred fields.
        #
        # We are making sure this doesn't trigger a scmtool_id or tool_id
        # lookup.
        #
        # 1 query:
        #
        # 1. Fetch id, name for repository.
        queries = [
            {
                'model': Repository,
                'where': Q(pk=repository.pk),
                'only_fields': {'id', 'name'},
            },
        ]

        with assert_queries(queries):
            repository = (
                Repository.objects.filter(pk=repository.pk)
                .only('pk', 'name')
                .get()
            )

        # This direct access will trigger such a lookup, but the value should
        # be None.
        #
        # 1 query:
        #
        # 1. Fetch scmtool_id for repository.
        queries = [
            {
                'model': Repository,
                'where': Q(pk=repository.pk),
                'only_fields': {'scmtool_id'},
            },
        ]

        with assert_queries(queries):
            self.assertIsNone(repository.scmtool_id)

    def test_scmtool_class_with_unset_scmtool_id_on_new_repo(self):
        """Testing Repository.scmtool_class with unset scmtool_id and with
        legacy tool_id on new (unsaved) repository
        """
        tool = Tool.objects.get(name='Git')

        with self.assertNumQueries(0):
            repository = Repository(tool=tool)

            self.assertIs(repository.scmtool_class, GitTool)
            self.assertEqual(repository.scmtool_id, 'git')
            self.assertIsNone(repository.pk)

    def test_scmtool_class_with_unset_scmtool_id_and_missing_scmtool(self):
        """Testing Repository.scmtool_class with unset scmtool_id and with
        legacy tool_id and missing SCMTool
        """
        from reviewboard.scmtools.signal_handlers import logger

        tool = Tool.objects.create(name='XXX', class_name='x.x.x.Tool')

        # 1 query:
        #
        # 1. Insert new repository.
        queries = [
            {
                'type': 'INSERT',
                'model': Repository,
            },
        ]

        with self.assertLogs(logger) as logs, assert_queries(queries):
            repository = Repository.objects.create(name='test-repo', tool=tool)

        self.assertEqual(logs.output, [
            'ERROR:reviewboard.scmtools.signal_handlers:Attempted to upgrade '
            'the SCMTool state for repository ID None, but the SCMTool "XXX" '
            'could not be loaded: Error importing SCM Tool x.x.x: '
            '"No module named \'x\'"',
        ])

        message = (
            'There was an error loading the SCMTool "XXX" needed by this '
            'repository. The administrator should ensure all necessary '
            'packages and extensions are installed.'
        )

        with self.assertNumQueries(0), \
             self.assertRaisesMessage(ImproperlyConfigured, message):
            repository.scmtool_class

        self.assertIsNone(repository.scmtool_id)

    def test_scmtool_class_with_unset_scmtool_id_and_no_tool(self):
        """Testing Repository.scmtool_class with scmtool_id and without
        legacy tool_id
        """
        with self.assertNumQueries(0):
            repository = Repository()

            self.assertIsNone(repository.scmtool_class)

    def test_scmtool_class_with_unregistered_scmtool_id(self):
        """Testing Repository.scmtool_class with unregistered SCMTool"""
        from reviewboard.scmtools.models import logger

        tool = Tool.objects.create(name='XXX', class_name='x.x.x.Tool')

        # 1 query:
        #
        # 1. Insert new repository.
        queries = [
            {
                'type': 'INSERT',
                'model': Repository,
            },
        ]

        with assert_queries(queries):
            repository = Repository.objects.create(name='test-repo',
                                                   tool=tool,
                                                   scmtool_id='xxxtool')

        message = (
            'There was an error loading the SCMTool "xxxtool" needed by this '
            'repository. The administrator should ensure all necessary '
            'packages and extensions are installed.'
        )

        with self.assertNumQueries(0):
            with self.assertRaisesMessage(ImproperlyConfigured, message):
                with self.assertLogs(logger) as logs:
                    repository.scmtool_class

        self.assertEqual(logs.output, [
            'ERROR:reviewboard.scmtools.models:Error finding registered '
            'SCMTool "xxxtool" in repository ID 2.',
        ])

    def test_password_decryption_failed(self) -> None:
        """Testing the repository password with failed decryption"""
        repository = self.repository
        repository.password = 'this is a super secret password'
        repository.save()

        repository.refresh_from_db()

        settings = {
            'SECRET_KEY': 'acdef12345acdef123456abcdef123456abcdef12345',
        }

        with self.settings(**settings), \
             self.assertLogs(level='CRITICAL') as logs:
            self.assertIsNone(repository.password)

        self.assertEqual(logs.output, [
            f'CRITICAL:reviewboard.scmtools.models:Unable to decrypt stored '
            f'password for repository pk={repository.pk}',
        ])



from django import forms
from django.utils.translation import gettext_lazy as _

from reviewboard.admin.validation import validate_bug_tracker_base_hosting_url
from reviewboard.hostingsvcs.base.forms import BaseHostingServiceRepositoryForm
from reviewboard.hostingsvcs.base.hosting_service import BaseHostingService


class VersionOneForm(BaseHostingServiceRepositoryForm):
    versionone_url = forms.CharField(
        label=_('VersionOne URL'),
        max_length=64,
        required=True,
        widget=forms.TextInput(attrs={'size': '60'}),
        validators=[validate_bug_tracker_base_hosting_url])


class VersionOne(BaseHostingService):
    name = 'VersionOne'
    hosting_service_id = 'versionone'
    form = VersionOneForm
    bug_tracker_field = '%(versionone_url)s/assetdetail.v1?Number=%%s'
    supports_bug_trackers = True



"""Unit tests for reviewboard.datagrids.columns.DiffUpdatedColumn.

Version Added:
    8.0
"""

from __future__ import annotations

from datetime import datetime, timezone

from django.utils import timezone as django_timezone

from reviewboard.datagrids.columns import DiffUpdatedColumn
from reviewboard.datagrids.tests.base import BaseColumnTestCase


class DiffUpdatedColumnTests(BaseColumnTestCase):
    """Unit tests for reviewboard.datagrids.columns.DiffUpdatedColumn.

    Version Added:
        8.0
    """

    column = DiffUpdatedColumn()
    fixtures = ['test_users', 'test_scmtools']

    def test_render_data_with_diff(self) -> None:
        """Testing DiffUpdatedColumn.render_data with a diff"""
        review_request = self.create_review_request(
            create_repository=True,
            publish=True)
        self.create_diffset(review_request)

        # This is normally set by augmenting the queryset when the diff is
        # added to the history.
        assert review_request.diffset_history is not None

        review_request.diffset_history.last_diff_updated = \
            datetime(2026, 1, 1, 12, 30, 40, tzinfo=timezone.utc)

        value = self.column.render_data(self.stateful_column, review_request)

        self.assertIsInstance(value, str)
        self.assertEqual(value, 'Jan. 1, 2026')

    def test_render_data_without_diff(self) -> None:
        """Testing DiffUpdatedColumn.render_data without a diff"""
        review_request = self.create_review_request(publish=True)

        # This is normally populated when a diff is added.
        review_request.diffset_history.last_diff_updated = None

        value = self.column.render_data(self.stateful_column, review_request)

        self.assertIsInstance(value, str)
        self.assertEqual(value, '')

    def test_to_json_with_diff(self) -> None:
        """Testing DiffUpdatedColumn.to_json with a diff"""
        review_request = self.create_review_request(
            create_repository=True,
            publish=True)
        self.create_diffset(review_request)

        # This is normally set by augmenting the queryset when the diff is
        # added to the history.
        review_request.diffset_history.last_diff_updated = \
            django_timezone.now()

        result = self.column.to_json(self.stateful_column, review_request)

        self.assertIsNotNone(result)

    def test_to_json_without_diff(self) -> None:
        """Testing DiffUpdatedColumn.to_json without a diff"""
        review_request = self.create_review_request(publish=True)

        # This is normally populated when a diff is added.
        review_request.diffset_history.last_diff_updated = None

        self.assertIsNone(
            self.column.to_json(self.stateful_column, review_request))



"""A backend for the Elasticsearch search engine."""

from django import forms
from django.core.exceptions import ValidationError
from django.utils.translation import gettext, gettext_lazy as _
from djblets.util.humanize import humanize_list

from reviewboard.search.search_backends.base import (SearchBackend,
                                                     SearchBackendForm)

try:
    import elasticsearch
    es_version = elasticsearch.VERSION
except Exception:
    es_version = None


# NOTE: When updating the versions below, make sure you've also updated
#       the documentation in docs/manual/admin/sites/search-indexing.rst.


#: The supported major versions of Elasticsearch.
#:
#: There must be a backend class available for each version in this list.
#:
#: Type:
#:     tuple
SUPPORTED_ES_MAJOR_VERSIONS = (7, 5, 2, 1)

#: The latest supported major version of Elasticsearch.
#:
#: Type:
#:     int
LATEST_ES_MAJOR_VERSION = SUPPORTED_ES_MAJOR_VERSIONS[0]

#: Whether the installed version of the elasticsearch module is supported.
#:
#: Type:
#:     bool
ES_VERSION_SUPPORTED = (es_version is not None and
                        es_version[0] in SUPPORTED_ES_MAJOR_VERSIONS)


# Determine the module name for the Haystack Elasticsearch backend.
if ES_VERSION_SUPPORTED and es_version[0] > 1:
    _engine_class_version = es_version[0]
else:
    _engine_class_version = ''


class ElasticsearchConfigForm(SearchBackendForm):
    """A form for configuring the Elasticsearch search backend."""

    url = forms.URLField(
        label=_('Elasticsearch URL'),
        help_text=_('The URL of the Elasticsearch server.'),
        widget=forms.TextInput(attrs={'size': 80}))

    index_name = forms.CharField(
        label=_('Elasticsearch index name'),
        help_text=_('The name of the Elasticsearch index.'),
        widget=forms.TextInput(attrs={'size': 40}))

    class Meta:
        if ES_VERSION_SUPPORTED:
            title = _('Elasticsearch %s.x') % es_version[0]
        else:
            title = _('Elasticsearch')

        fieldsets = (
            (None, {
                'description': _(
                    'Elasticsearch support requires a version of the '
                    '<a href="%(elasticsearch_url)s" target="_blank">'
                    'elasticsearch</a> Python package that both matches your '
                    'version of the Elasticsearch server and is compatible '
                    'with Review Board.'
                    '\n'
                    'We provide convenient packages for each supported major '
                    'version of Elasticsearch. For example, for Elasticsearch '
                    '%(major_version)s.x, install '
                    '<code>ReviewBoard[elasticsearch%(major_version)s]</code> '
                    'and then restart your web server:'
                    '\n'
                    '<code><strong>$</strong> pip install '
                    '"ReviewBoard[elasticsearch%(major_version)s]"</code>'
                    '\n'
                    '%(supported_versions)s are supported.'
                ) % {
                    'elasticsearch_url':
                        'https://pypi.org/project/elasticsearch/',
                    'major_version': LATEST_ES_MAJOR_VERSION,
                    'supported_versions': humanize_list([
                        '%s.x' % _major_version
                        for _major_version in SUPPORTED_ES_MAJOR_VERSIONS
                    ]),
                },
                'fields': ('url', 'index_name'),
            }),
        )


class ElasticsearchBackend(SearchBackend):
    """A search backend for integrating with Elasticsearch"""

    search_backend_id = 'elasticsearch'
    name = _('Elasticsearch')
    haystack_backend_name = (
        f'haystack.backends.'
        f'elasticsearch{_engine_class_version}_backend.'
        f'Elasticsearch{_engine_class_version}SearchEngine'
    )
    default_settings = {
        'URL': 'http://127.0.0.1:9200/',
        'INDEX_NAME': 'reviewboard',
    }
    config_form_class = ElasticsearchConfigForm
    form_field_map = {
        'url': 'URL',
        'index_name': 'INDEX_NAME',
    }

    # Assign these here to allow for unit tests to override them.
    _ES_VERSION_SUPPORTED = ES_VERSION_SUPPORTED
    _es_version = es_version

    def validate(
        self,
        **kwargs,
    ) -> None:
        """Ensure that the elasticsearch Python module is installed.

        Raises:
            django.core.exceptions.ValidationError:
                Raised if the ``elasticsearch`` module is not installed or
                the version is incompatible.
        """
        # Check whether there's a supported version of the module available.
        # Note that technically, elasticsearch 1.x is supported, but it's
        # pretty old. If we're going to reference a version, we want to
        # reference 2.x.
        if not self._ES_VERSION_SUPPORTED:
            es_version = self._es_version
            supported_packages = ', '.join(
                'ReviewBoard[elasticsearch%s]' % _version
                for _version in SUPPORTED_ES_MAJOR_VERSIONS
            )

            if es_version:
                raise ValidationError(
                    gettext('You need to install a supported version of the '
                            'elasticsearch module. Version %(cur_version)s '
                            'is not supported. Compatible packages are: '
                            '%(packages)s')
                    % {
                        'cur_version': '%s.%s' % (es_version[0],
                                                  es_version[1]),
                        'packages': supported_packages,
                    })
            else:
                raise ValidationError(
                    gettext('You need to install a supported version of the '
                            'elasticsearch module. Compatible packages are: '
                            '%(packages)s')
                    % {
                        'packages': supported_packages,
                    })

        super(ElasticsearchBackend, self).validate(**kwargs)



"""Unit tests for reviewboard.certs.http.CertificateVerificationHTTPSHandler.

Version Added:
    8.1
"""

from __future__ import annotations

import os
import ssl
from urllib.request import Request

import kgb

from reviewboard.certs.cert import CertPurpose, Certificate, CertificateBundle
from reviewboard.certs.errors import CertificateVerificationError
from reviewboard.certs.http import CertificateVerificationHTTPSHandler
from reviewboard.certs.manager import CertificateManager
from reviewboard.certs.tests.testcases import (CaptureSSLContext,
                                               TEST_CERT_BUNDLE_PEM,
                                               TEST_CLIENT_CERT_PEM,
                                               TEST_CLIENT_KEY_PEM,
                                               TEST_TRUST_CERT_PEM)
from reviewboard.testing.testcase import TestCase


class CertificateVerificationHTTPSHandlerTests(kgb.SpyAgency, TestCase):
    """Unit tests for CertificateVerificationHTTPSHandler.

    Version Added:
        8.1
    """

    def setUp(self) -> None:
        """Set up state for the test."""
        super().setUp()

        self.spy_on(
            Certificate.create_from_server,
            owner=Certificate,
            op=kgb.SpyOpReturn(Certificate(
                hostname='example.com',
                port=443,
                cert_data=TEST_TRUST_CERT_PEM,
            )),
        )

    def test_https_open_with_cert_found(self) -> None:
        """Testing CertificateVerificationHTTPSHandler.https_open
        with certificate found for hostname
        """
        cert_manager = CertificateManager()
        cert_manager.add_ca_bundle(CertificateBundle(
            bundle_data=TEST_CERT_BUNDLE_PEM,
            name='my-certs',
        ))
        cert_manager.add_certificate(self.create_certificate(
            cert_data=TEST_TRUST_CERT_PEM,
        ))
        cert_manager.add_certificate(self.create_certificate(
            purpose=CertPurpose.CLIENT,
            cert_data=TEST_CLIENT_CERT_PEM,
            key_data=TEST_CLIENT_KEY_PEM,
        ))

        storage_backend = cert_manager.storage_backend

        opener = CertificateVerificationHTTPSHandler(
            cert_manager=cert_manager,
            local_site=None,
        )

        self.spy_on(ssl.create_default_context,
                    op=kgb.SpyOpReturn(CaptureSSLContext()))
        self.spy_on(opener.do_open, call_original=False)

        opener.https_open(Request(url='https://example.com'))
        context = opener._context

        assert isinstance(context, CaptureSSLContext)
        self.assertAttrsEqual(
            context,
            {
                'cadatas': [],
                'cafiles': [
                    os.path.join(storage_backend.storage_path, 'certs',
                                 'trust', 'example.com__443.crt'),
                ],
                'capaths': [
                    os.path.join(storage_backend.storage_path, 'cabundles'),
                ],
                'certfiles': [
                    os.path.join(storage_backend.storage_path, 'certs',
                                 'client', 'example.com__443.crt'),
                ],
                'keyfiles': [
                    os.path.join(storage_backend.storage_path, 'certs',
                                 'client', 'example.com__443.key'),
                ],
                'passwords': [
                    None,
                ],
            })

    def test_https_open_with_cert_not_found(self) -> None:
        """Testing CertificateVerificationHTTPSHandler.https_open
        with certificate not found for hostname
        """
        cert_manager = CertificateManager()
        cert_manager.add_ca_bundle(CertificateBundle(
            bundle_data=TEST_CERT_BUNDLE_PEM,
            name='my-certs',
        ))
        cert_manager.add_certificate(self.create_certificate(
            cert_data=TEST_TRUST_CERT_PEM,
        ))
        cert_manager.add_certificate(self.create_certificate(
            purpose=CertPurpose.CLIENT,
            cert_data=TEST_CLIENT_CERT_PEM,
            key_data=TEST_CLIENT_KEY_PEM,
        ))

        storage_backend = cert_manager.storage_backend

        opener = CertificateVerificationHTTPSHandler(
            cert_manager=cert_manager,
            local_site=None,
        )

        self.spy_on(ssl.create_default_context,
                    op=kgb.SpyOpReturn(CaptureSSLContext()))
        self.spy_on(opener.do_open, call_original=False)

        opener.https_open(Request(url='https://other.example.com'))
        context = opener._context

        assert isinstance(context, CaptureSSLContext)
        self.assertAttrsEqual(
            context,
            {
                'cadatas': [],
                'cafiles': [],
                'capaths': [
                    os.path.join(storage_backend.storage_path, 'cabundles'),
                ],
                'certfiles': [],
                'keyfiles': [],
                'passwords': [],
            })

    def test_process_ssl_error_with_not_yet_valid(self) -> None:
        """Testing CertificateVerificationHTTPSHandler._process_ssl_error
        with X509_V_ERR_CERT_NOT_YET_VALID
        """
        self._run_process_cert_verification_error_test(
            verify_code=9,
            verify_code_name='X509_V_ERR_CERT_NOT_YET_VALID',
            expected_message=(
                'The SSL certificate provided by example.com is not yet '
                'valid and cannot be used.'
            ),
        )

    def test_process_ssl_error_with_expired(self) -> None:
        """Testing CertificateVerificationHTTPSHandler._process_ssl_error
        with X509_V_ERR_CERT_HAS_EXPIRED
        """
        self._run_process_cert_verification_error_test(
            verify_code=10,
            verify_code_name='X509_V_ERR_CERT_HAS_EXPIRED',
            expected_message=(
                'The SSL certificate provided by example.com has expired and '
                'can no longer be used.'
            ),
        )

    def test_process_ssl_error_with_depth_zero_self_signed_cert(self) -> None:
        """Testing CertificateVerificationHTTPSHandler._process_ssl_error
        with X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT
        """
        self._run_process_cert_verification_error_test(
            verify_code=18,
            verify_code_name='X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT',
            expected_message=(
                'The SSL certificate provided by example.com has not been '
                'signed by a trusted Certificate Authority and may not be '
                'safe. The certificate needs to be verified in Review Board '
                'before the server can be accessed.'
            ),
        )

    def test_process_ssl_error_with_self_signed_cert_in_chain(self) -> None:
        """Testing CertificateVerificationHTTPSHandler._process_ssl_error
        with X509_V_ERR_SELF_SIGNED_CERT_IN_CHAIN
        """
        self._run_process_cert_verification_error_test(
            verify_code=19,
            verify_code_name='X509_V_ERR_SELF_SIGNED_CERT_IN_CHAIN',
            expected_message=(
                'The SSL certificate provided by example.com has not been '
                'signed by a trusted Certificate Authority and may not be '
                'safe. The certificate needs to be verified in Review Board '
                'before the server can be accessed.'
            ),
        )

    def test_process_ssl_error_with_unable_to_get_issuer_cert_locally(
        self,
    ) -> None:
        """Testing CertificateVerificationHTTPSHandler._process_ssl_error
        with X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY
        """
        self._run_process_cert_verification_error_test(
            verify_code=20,
            verify_code_name='X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY',
            expected_message=(
                'The SSL certificate provided by example.com has not been '
                'signed by a trusted Certificate Authority and may not be '
                'safe. The certificate needs to be verified in Review Board '
                'before the server can be accessed.'
            ),
        )

    def test_process_ssl_error_with_unable_to_verify_leaf_signature(
        self,
    ) -> None:
        """Testing CertificateVerificationHTTPSHandler._process_ssl_error
        with X509_V_ERR_UNABLE_TO_VERIFY_LEAF_SIGNATURE
        """
        self._run_process_cert_verification_error_test(
            verify_code=21,
            verify_code_name='X509_V_ERR_UNABLE_TO_VERIFY_LEAF_SIGNATURE',
            expected_message=(
                'The SSL certificate provided by example.com has not been '
                'signed by a trusted Certificate Authority and may not be '
                'safe. The certificate needs to be verified in Review Board '
                'before the server can be accessed.'
            ),
        )

    def test_process_ssl_error_with_cert_untrusted(self) -> None:
        """Testing CertificateVerificationHTTPSHandler._process_ssl_error
        with X509_V_ERR_CERT_UNTRUSTED
        """
        self._run_process_cert_verification_error_test(
            verify_code=27,
            verify_code_name='X509_V_ERR_CERT_UNTRUSTED',
            expected_message=(
                'The SSL certificate provided by example.com has not been '
                'signed by a trusted Certificate Authority and may not be '
                'safe. The certificate needs to be verified in Review Board '
                'before the server can be accessed.'
            ),
        )

    def test_process_ssl_error_with_hostname_mismatch(self) -> None:
        """Testing CertificateVerificationHTTPSHandler._process_ssl_error
        with X509_V_ERR_HOSTNAME_MISMATCH
        """
        self._run_process_cert_verification_error_test(
            verify_code=62,
            verify_code_name='X509_V_ERR_HOSTNAME_MISMATCH',
            expected_message=(
                'The SSL certificate provided by example.com does not '
                'match its hostname and may not be safe.'
            ),
        )

    def test_process_ssl_error_with_other_cert_code(self) -> None:
        """Testing CertificateVerificationHTTPSHandler._process_ssl_error
        with other cert code
        """
        self._run_process_cert_verification_error_test(
            verify_code=1234567,
            verify_code_name='Some Other Bad Thing Happened',
            expected_message=(
                'The SSL certificate provided by example.com could not be '
                'verified and may not be safe. The certificate must be valid '
                'and verified in Review Board before the server can be '
                'accessed.'
            ),
        )

    def test_process_ssl_error_with_sslerror(self) -> None:
        """Testing CertificateVerificationHTTPSHandler._process_ssl_error
        with SSLError
        """
        opener = CertificateVerificationHTTPSHandler(local_site=None)

        error = ssl.SSLError('Bob turned traitor')
        message = (
            'The SSL certificate provided by example.com could not be '
            'verified and may not be safe. The certificate must be valid '
            'and verified in Review Board before the server can be accessed. '
            'Certificate details: hostname="example.com", port=443, '
            'issuer="example.com", fingerprints=SHA1=F2:35:0F:BB:34:40:84:'
            '78:8B:20:1D:40:B1:4A:17:0C:DE:36:2F:D5; SHA256=79:19:70:AE:A6:'
            '1B:EB:BC:35:7C:B8:54:B1:6A:AD:79:FF:F7:28:69:02:5E:C3:6F:B3:C2:'
            'B4:FD:84:66:DF:8F'
        )

        with self.assertRaisesMessage(CertificateVerificationError, message):
            opener._process_ssl_error(error=error,
                                      hostname='example.com',
                                      port=443)

    def _run_process_cert_verification_error_test(
        self,
        *,
        verify_code: int,
        verify_code_name: str,
        expected_message: str,
    ) -> None:
        """Run an SSL cert verification error test.

        Args:
            verify_code (int):
                The SSL verification error code.

            verify_code_name (str):
                The SSL verification error name string.

            expected_message (str):
                The expected SSL error message.
        """
        opener = CertificateVerificationHTTPSHandler(local_site=None)

        error = ssl.SSLCertVerificationError()
        error.verify_code = verify_code
        error.verify_message = verify_code_name

        message = (
            f'{expected_message} '
            f'Certificate details: hostname="example.com", port=443, '
            f'issuer="example.com", fingerprints=SHA1=F2:35:0F:BB:34:40:84:'
            f'78:8B:20:1D:40:B1:4A:17:0C:DE:36:2F:D5; SHA256=79:19:70:AE:A6:'
            f'1B:EB:BC:35:7C:B8:54:B1:6A:AD:79:FF:F7:28:69:02:5E:C3:6F:B3:C2:'
            f'B4:FD:84:66:DF:8F'
        )

        with self.assertRaisesMessage(CertificateVerificationError, message):
            opener._process_ssl_error(error=error,
                                      hostname='example.com',
                                      port=443)



"""Unit tests for BaseReviewRequestDetails."""

from __future__ import annotations

from django_assert_queries.testing import assert_queries
from django.contrib.auth.models import User
from django.db.models import Max, Q
from djblets.testing.decorators import add_fixtures

from reviewboard.attachments.models import (
    FileAttachment,
    FileAttachmentHistory)
from reviewboard.reviews.models import DefaultReviewer, ReviewRequest
from reviewboard.testing import TestCase


class BaseReviewRequestDetailsTests(TestCase):
    """Unit tests for BaseReviewRequestDetails."""

    fixtures = ['test_scmtools']

    def test_add_default_reviewers_with_users(self):
        """Testing BaseReviewRequestDetails.add_default_reviewers with users"""
        user1 = User.objects.create_user(username='user1',
                                         email='user1@example.com')
        user2 = User.objects.create(username='user2',
                                    email='user2@example.com')
        user3 = User.objects.create(username='user3',
                                    email='user3@example.com',
                                    is_active=False)

        # Create the default reviewers.
        default_reviewer1 = DefaultReviewer.objects.create(name='Test 1',
                                                           file_regex='.*')
        default_reviewer1.people.add(user1, user2, user3)

        # Create the review request and accompanying diff.
        review_request = self.create_review_request(create_repository=True,
                                                    submitter=user1)
        diffset = self.create_diffset(review_request)
        self.create_filediff(diffset)

        # The following queries will be executed:
        #
        # 1. Diffset
        # 2. The default reviewer list
        # 3. The file list
        # 4. User list for all matched default reviewers
        # 5. Update users (m2m.add())
        # 6. Group list for all matched default reviewers
        with self.assertNumQueries(6):
            review_request.add_default_reviewers()

        self.assertEqual(list(review_request.target_people.all()),
                         [user1, user2])
        self.assertEqual(list(review_request.target_groups.all()),
                         [])

    def test_add_default_reviewers_with_groups(self):
        """Testing BaseReviewRequestDetails.add_default_reviewers with groups
        """
        user1 = User.objects.create_user(username='user1',
                                         email='user1@example.com')

        group1 = self.create_review_group(name='Group 1')
        group2 = self.create_review_group(name='Group 2')

        # Create the default reviewers.
        default_reviewer1 = DefaultReviewer.objects.create(name='Test 1',
                                                           file_regex='.*')
        default_reviewer1.groups.add(group1, group2)

        # Create the review request and accompanying diff.
        review_request = self.create_review_request(create_repository=True,
                                                    submitter=user1)
        diffset = self.create_diffset(review_request)
        self.create_filediff(diffset)

        # The following queries will be executed:
        #
        # 1. Diffset
        # 2. The default reviewer list
        # 3. The file list
        # 4. User list for all matched default reviewers
        # 5. Group list for all matched default reviewers
        # 6. Update groups (m2m.add())
        with self.assertNumQueries(6):
            review_request.add_default_reviewers()

        self.assertEqual(list(review_request.target_groups.all()),
                         [group1, group2])
        self.assertEqual(list(review_request.target_people.all()),
                         [])

    def test_add_default_reviewers_with_users_and_groups(self):
        """Testing BaseReviewRequestDetails.add_default_reviewers with both
        users and groups
        """
        user1 = User.objects.create_user(username='user1',
                                         email='user1@example.com')
        user2 = User.objects.create(username='user2',
                                    email='user2@example.com')

        group1 = self.create_review_group(name='Group 1')
        group2 = self.create_review_group(name='Group 2')

        # Create the default reviewers.
        default_reviewer1 = DefaultReviewer.objects.create(name='Test 1',
                                                           file_regex='.*')
        default_reviewer1.people.add(user1, user2)

        default_reviewer2 = DefaultReviewer.objects.create(name='Test 2',
                                                           file_regex='.*')
        default_reviewer2.groups.add(group1, group2)

        # Create the review request and accompanying diff.
        review_request = self.create_review_request(create_repository=True,
                                                    submitter=user1)
        diffset = self.create_diffset(review_request)
        self.create_filediff(diffset)

        # The following queries will be executed:
        #
        # 1. Diffset
        # 2. The default reviewer list
        # 3. The file list
        # 4. User list for all matched default reviewers
        # 5. Update users (m2m.add())
        # 6. Group list for all matched default reviewers
        # 7. Update groups (m2m.add())
        with self.assertNumQueries(7):
            review_request.add_default_reviewers()

        self.assertEqual(list(review_request.target_people.all()),
                         [user1, user2])
        self.assertEqual(list(review_request.target_groups.all()),
                         [group1, group2])

    def test_add_default_reviewers_with_no_matches(self):
        """Testing BaseReviewRequestDetails.add_default_reviewers with no
        matches
        """
        user1 = User.objects.create_user(username='user1',
                                         email='user1@example.com')
        user2 = User.objects.create(username='user2',
                                    email='user2@example.com')

        group1 = self.create_review_group(name='Group 1')
        group2 = self.create_review_group(name='Group 2')

        # Create the default reviewers.
        default_reviewer1 = DefaultReviewer.objects.create(name='Test 1',
                                                           file_regex='/foo')
        default_reviewer1.people.add(user1, user2)

        default_reviewer2 = DefaultReviewer.objects.create(name='Test 2',
                                                           file_regex='/bar')
        default_reviewer2.groups.add(group1, group2)

        # Create the review request and accompanying diff.
        review_request = self.create_review_request(create_repository=True,
                                                    submitter=user1)
        diffset = self.create_diffset(review_request)
        self.create_filediff(diffset)

        # The following queries will be executed:
        #
        # 1. Diffset
        # 2. The file list
        # 3. The default reviewer list
        with self.assertNumQueries(3):
            review_request.add_default_reviewers()

        self.assertEqual(list(review_request.target_people.all()), [])
        self.assertEqual(list(review_request.target_groups.all()), [])

    def test_add_default_reviewers_with_no_repository(self):
        """Testing BaseReviewRequestDetails.add_default_reviewers with no
        repository
        """
        user1 = User.objects.create_user(username='user1',
                                         email='user1@example.com')

        # Create the review request and accompanying diff.
        review_request = self.create_review_request(submitter=user1)

        with self.assertNumQueries(0):
            review_request.add_default_reviewers()

    @add_fixtures(['test_users'])
    def test_get_file_attachments(self) -> None:
        """Testing BaseReviewRequestDetails.get_file_attachments"""
        review_request = self.create_review_request()

        active = self.create_file_attachment(review_request)
        active_2 = self.create_file_attachment(review_request)

        self.create_file_attachment(review_request, active=False)

        # 3 queries:
        #
        # 1. Fetch active file attachments.
        # 2. Fetch display position for the first attachment
        # 3. Fetch display position for the second attachment
        queries = [
            {
                'join_types': {
                    'reviews_reviewrequest_file_attachments': 'INNER JOIN',
                },
                'model': FileAttachment,
                'num_joins': 1,
                'tables': {
                    'attachments_fileattachment',
                    'reviews_reviewrequest_file_attachments',
                },
                'where': Q(review_request__id=review_request.pk),
            },
            {
                'model': FileAttachmentHistory,
                'where': Q(id=active.pk),
            },
            {
                'model': FileAttachmentHistory,
                'where': Q(id=active_2.pk),
            },
        ]

        with assert_queries(queries):
            attachments = review_request.get_file_attachments()

        self.assertListEqual(attachments, [active, active_2])

    @add_fixtures(['test_users'])
    def test_get_file_attachments_with_sort_false(self) -> None:
        """Testing BaseReviewRequestDetails.get_file_attachments with
        sort=False
        """
        review_request = self.create_review_request()

        active = self.create_file_attachment(review_request)
        active_2 = self.create_file_attachment(review_request)

        self.create_file_attachment(review_request, active=False)

        # 1 query:
        #
        # 1. Fetch active file attachments.
        queries = [
            {
                'join_types': {
                    'reviews_reviewrequest_file_attachments': 'INNER JOIN',
                },
                'model': FileAttachment,
                'num_joins': 1,
                'tables': {
                    'attachments_fileattachment',
                    'reviews_reviewrequest_file_attachments',
                },
                'where': Q(review_request__id=review_request.pk),
            },
        ]

        with assert_queries(queries):
            attachments = review_request.get_file_attachments(sort=False)

        self.assertListEqual(attachments, [active, active_2])

    @add_fixtures(['test_users'])
    def test_get_file_attachments_legacy(self) -> None:
        """Testing BaseReviewRequestDetails.get_file_attachments with legacy
        attachments that don't have an associated FileAttachmentHistory
        """
        review_request = self.create_review_request()

        active = self.create_file_attachment(
            review_request,
            with_history=False)
        active_2 = self.create_file_attachment(
            review_request,
            with_history=False)

        self.create_file_attachment(review_request, active=False)

        # 9 queries:
        #
        #   1. Fetch active file attachments.
        # 2-5. Create a FileAttachmentHistory for the first attachment.
        # 6-9. Create a FileAttachmentHistory for the second attachment.
        queries = [
            {
                'join_types': {
                    'reviews_reviewrequest_file_attachments': 'INNER JOIN',
                },
                'model': FileAttachment,
                'num_joins': 1,
                'tables': {
                    'attachments_fileattachment',
                    'reviews_reviewrequest_file_attachments',
                },
                'where': Q(review_request__id=review_request.pk),
            },
            {
                'join_types': {
                    'reviews_reviewrequest_file_attachment_histories':
                        'INNER JOIN',
                },
                'model': FileAttachmentHistory,
                'annotations': {
                    'display_position__max': Max('display_position'),
                },
                'num_joins': 1,
                'tables': {
                    'attachments_fileattachmenthistory',
                    'reviews_reviewrequest_file_attachment_histories',
                },
                'where': Q(review_request=review_request),
            },
            {
                'model': FileAttachmentHistory,
                'type': 'INSERT',
            },
            {
                'model': ReviewRequest.file_attachment_histories.through,
                'type': 'INSERT',
            },
            {
                'model': FileAttachment,
                'type': 'UPDATE',
                'where': Q(pk=active.pk)
            },
            {
                'join_types': {
                    'reviews_reviewrequest_file_attachment_histories':
                        'INNER JOIN',
                },
                'model': FileAttachmentHistory,
                'annotations': {
                    'display_position__max': Max('display_position'),
                },
                'num_joins': 1,
                'tables': {
                    'attachments_fileattachmenthistory',
                    'reviews_reviewrequest_file_attachment_histories',
                },
                'where': Q(review_request=review_request),
            },
            {
                'model': FileAttachmentHistory,
                'type': 'INSERT',
            },
            {
                'model': ReviewRequest.file_attachment_histories.through,
                'type': 'INSERT',
            },
            {
                'model': FileAttachment,
                'type': 'UPDATE',
                'where': Q(pk=active_2.pk)
            },
        ]

        with assert_queries(queries):
            attachments = review_request.get_file_attachments()

        self.assertListEqual(attachments, [active, active_2])

    @add_fixtures(['test_users'])
    def test_get_file_attachments_with_no_attachments(self) -> None:
        """Testing BaseReviewRequestDetails.get_file_attachments with
        no active attachments
        """
        review_request = self.create_review_request()

        self.create_file_attachment(review_request, active=False)

        with self.assertNumQueries(0):
            attachments = review_request.get_file_attachments(sort=False)

        self.assertListEqual(attachments, [])



import os

from django.contrib.auth.models import AnonymousUser, User

from reviewboard.reviews.models import Group
from reviewboard.scmtools.forms import RepositoryForm
from reviewboard.scmtools.models import Repository
from reviewboard.site.models import LocalSite
from reviewboard.testing.testcase import TestCase


class PolicyTests(TestCase):
    """Unit tests for access policies."""

    fixtures = ['test_scmtools']

    def setUp(self):
        self.user = User.objects.create_user(username='testuser', password='',
                                             email='user@example.com')
        self.anonymous = AnonymousUser()
        self.repo = self.create_repository(
            name='test',
            path='example.com:/cvsroot/test',
            username='anonymous',
            tool_name='CVS')

    def test_repository_public(self):
        """Testing access to a public repository"""
        self.assertTrue(self.repo.is_accessible_by(self.user))
        self.assertTrue(self.repo.is_accessible_by(self.anonymous))

        self.assertIn(self.repo, Repository.objects.accessible(self.user))
        self.assertTrue(
            self.repo in Repository.objects.accessible(self.anonymous))

    def test_repository_private_access_denied(self):
        """Testing no access to an inaccessible private repository"""
        self.repo.public = False
        self.repo.save()

        self.assertFalse(self.repo.is_accessible_by(self.user))
        self.assertFalse(self.repo.is_accessible_by(self.anonymous))

        self.assertNotIn(self.repo, Repository.objects.accessible(self.user))
        self.assertFalse(
            self.repo in Repository.objects.accessible(self.anonymous))

    def test_repository_private_access_allowed_by_user(self):
        """Testing access to a private repository accessible by user"""
        self.repo.users.add(self.user)
        self.repo.public = False
        self.repo.save()

        self.assertTrue(self.repo.is_accessible_by(self.user))
        self.assertFalse(self.repo.is_accessible_by(self.anonymous))

        self.assertIn(self.repo, Repository.objects.accessible(self.user))
        self.assertFalse(
            self.repo in Repository.objects.accessible(self.anonymous))

    def test_repository_private_access_allowed_by_review_group(self):
        """Testing access to a private repository accessible by review group"""
        group = Group.objects.create(name='test-group')
        group.users.add(self.user)

        self.repo.public = False
        self.repo.review_groups.add(group)
        self.repo.save()

        self.assertTrue(self.repo.is_accessible_by(self.user))
        self.assertFalse(self.repo.is_accessible_by(self.anonymous))

        self.assertIn(self.repo, Repository.objects.accessible(self.user))
        self.assertFalse(
            self.repo in Repository.objects.accessible(self.anonymous))

    def test_repository_form_with_local_site_and_bad_group(self):
        """Testing adding a Group to a RepositoryForm with the wrong LocalSite
        """
        test_site = LocalSite.objects.create(name='test')
        group = Group.objects.create(name='test-group')

        svn_repo_path = 'file://' + os.path.join(os.path.dirname(__file__),
                                                 '..', 'testdata', 'svn_repo')

        form = RepositoryForm({
            'name': 'test',
            'path': svn_repo_path,
            'hosting_type': 'custom',
            'bug_tracker_type': 'custom',
            'review_groups': [group.pk],
            'local_site': test_site.pk,
            'tool': 'subversion',
        })
        self.assertFalse(form.is_valid())

        group.local_site = test_site
        group.save()

        form = RepositoryForm({
            'name': 'test',
            'path': svn_repo_path,
            'hosting_type': 'custom',
            'bug_tracker_type': 'custom',
            'review_groups': [group.pk],
            'tool': 'subversion',
        })
        self.assertFalse(form.is_valid())

    def test_repository_form_with_local_site_and_bad_user(self):
        """Testing adding a User to a RepositoryForm with the wrong LocalSite
        """
        test_site = LocalSite.objects.create(name='test')

        svn_repo_path = 'file://' + os.path.join(os.path.dirname(__file__),
                                                 '..', 'testdata', 'svn_repo')

        form = RepositoryForm({
            'name': 'test',
            'path': svn_repo_path,
            'hosting_type': 'custom',
            'bug_tracker_type': 'custom',
            'users': [self.user.pk],
            'local_site': test_site.pk,
            'tool': 'subversion',
        })
        self.assertFalse(form.is_valid())



from django.dispatch import Signal


#: Emitted when a user is added to a local site.
#:
#: Args:
#:     user (django.contrib.auth.models.User):
#:         The user which was added.
#:
#:     localsite (reviewboard.site.models.LocalSite):
#:         The site the user was added to.
local_site_user_added = Signal()



"""API resource for review request drafts."""

from __future__ import annotations

import logging
import re
from typing import TYPE_CHECKING

from django.contrib import auth
from django.contrib.auth.models import User
from django.core.exceptions import PermissionDenied
from django.db.models import ManyToManyField, Q
from djblets.util.decorators import augment_method_from
from djblets.webapi.decorators import (webapi_login_required,
                                       webapi_response_errors,
                                       webapi_request_fields)
from djblets.webapi.errors import (DOES_NOT_EXIST,
                                   INVALID_FORM_DATA,
                                   NOT_LOGGED_IN,
                                   PERMISSION_DENIED)
from djblets.webapi.fields import (BooleanFieldType,
                                   ChoiceFieldType,
                                   DateTimeFieldType,
                                   DictFieldType,
                                   IntFieldType,
                                   ResourceFieldType,
                                   ResourceListFieldType,
                                   StringFieldType)

from reviewboard.certs.errors import CertificateVerificationError
from reviewboard.diffviewer.features import dvcs_feature
from reviewboard.hostingsvcs.errors import HostingServiceError
from reviewboard.reviews.builtin_fields import BuiltinFieldMixin
from reviewboard.reviews.errors import NotModifiedError, PublishError
from reviewboard.reviews.fields import (get_review_request_fields,
                                        get_review_request_field)
from reviewboard.reviews.models import Group, ReviewRequest, ReviewRequestDraft
from reviewboard.scmtools.errors import (InvalidChangeNumberError,
                                         SCMError)
from reviewboard.webapi.base import ImportExtraDataError, WebAPIResource
from reviewboard.webapi.decorators import webapi_check_local_site
from reviewboard.webapi.errors import (COMMIT_ID_ALREADY_EXISTS,
                                       INVALID_CHANGE_NUMBER,
                                       NOTHING_TO_PUBLISH,
                                       PUBLISH_ERROR,
                                       REPO_INFO_ERROR,
                                       UNVERIFIED_HOST_CERT)
from reviewboard.webapi.mixins import MarkdownFieldsMixin
from reviewboard.webapi.resources import resources

if TYPE_CHECKING:
    from django.db.models import QuerySet
    from django.http import HttpRequest


logger = logging.getLogger(__name__)


class ReviewRequestDraftResource(MarkdownFieldsMixin, WebAPIResource):
    """An editable draft of a review request.

    This resource is used to actually modify a review request. Anything made
    in this draft can be published in order to become part of the public
    review request, or it can be discarded.

    Any POST or PUTs on this draft will cause the draft to be created
    automatically. An initial POST is not required.

    There is only ever a maximum of one draft per review request.

    In order to access this resource, the user must either own the review
    request, or it must have the ``reviews.can_edit_reviewrequest`` permission
    set.
    """

    model = ReviewRequestDraft
    name = 'draft'
    policy_id = 'review_request_draft'
    singleton = True
    model_parent_key = 'review_request'
    mimetype_item_resource_name = 'review-request-draft'
    uri_template_name = 'review_request_draft'

    fields = {
        'id': {
            'type': IntFieldType,
            'description': 'The numeric ID of the draft.',
        },
        'review_request': {
            'type': ResourceFieldType,
            'resource': 'reviewboard.webapi.resources.review_request.'
                        'ReviewRequestResource',
            'description': 'The review request that owns this draft.',
        },
        'last_updated': {
            'type': DateTimeFieldType,
            'description': 'The date and time that the draft was last '
                           'updated.',
        },
        'branch': {
            'type': StringFieldType,
            'description': 'The branch name.',
        },
        'bugs_closed': {
            'type': StringFieldType,
            'description': 'The new list of bugs closed or referenced by this '
                           'change.',
        },
        'depends_on': {
            'type': ResourceListFieldType,
            'resource': 'reviewboard.webapi.resources.review_request.'
                        'ReviewRequestResource',
            'description': 'The list of review requests that this '
                           'review request depends on.',
            'added_in': '1.7.8',
        },
        'changedescription': {
            'type': StringFieldType,
            'description': 'A custom description of what changes are being '
                           'made in this update. It often will be used to '
                           'describe the changes in the diff.',
            'supports_text_types': True,
        },
        'changedescription_text_type': {
            'type': ChoiceFieldType,
            'choices': MarkdownFieldsMixin.TEXT_TYPES,
            'description': 'The current or forced text type for the '
                           '``changedescription`` field.',
            'added_in': '2.0.12',
        },
        'commit_id': {
            'type': StringFieldType,
            'description': 'The updated ID of the commit this review request '
                           'is based upon.',
            'added_in': '2.0',
        },
        'description': {
            'type': StringFieldType,
            'description': 'The new review request description.',
            'supports_text_types': True,
        },
        'description_text_type': {
            'type': ChoiceFieldType,
            'choices': MarkdownFieldsMixin.TEXT_TYPES,
            'description': 'The current or forced text type for the '
                           '``description`` field.',
            'added_in': '2.0.12',
        },
        'extra_data': {
            'type': DictFieldType,
            'description': 'Extra data as part of the draft. '
                           'This can be set by the API or extensions.',
            'added_in': '2.0',
        },
        'public': {
            'type': BooleanFieldType,
            'description': 'Whether or not the draft is public. '
                           'This will always be false up until the time '
                           'it is first made public. At that point, the '
                           'draft is deleted.',
        },
        'submitter': {
            'type': StringFieldType,
            'description': 'The user who submitted the review request.',
        },
        'summary': {
            'type': StringFieldType,
            'description': 'The new review request summary.',
        },
        'target_groups': {
            'type': StringFieldType,
            'description': 'A comma-separated list of review groups '
                           'that will be on the reviewer list.',
        },
        'target_people': {
            'type': StringFieldType,
            'description': 'A comma-separated list of users that will '
                           'be on a reviewer list.',
        },
        'testing_done': {
            'type': StringFieldType,
            'description': 'The new testing done text.',
            'supports_text_types': True,
        },
        'testing_done_text_type': {
            'type': ChoiceFieldType,
            'choices': MarkdownFieldsMixin.TEXT_TYPES,
            'description': 'The current or forced text type for the '
                           '``testing_done`` field.',
            'added_in': '2.0.12',
        },
        'text_type': {
            'type': ChoiceFieldType,
            'choices': MarkdownFieldsMixin.TEXT_TYPES,
            'description': 'Formerly responsible for indicating the text '
                           'type for text fields. Replaced by '
                           '``changedescription_text_type``, '
                           '``description_text_type``, and '
                           '``testing_done_text_type`` in 2.0.12.',
            'added_in': '2.0',
            'deprecated_in': '2.0.12',
        },
    }

    allowed_methods = ('GET', 'POST', 'PUT', 'DELETE')

    item_child_resources = [
        resources.draft_diff,
        resources.draft_screenshot,
        resources.draft_file_attachment,
    ]

    CREATE_UPDATE_OPTIONAL_FIELDS = {
        'branch': {
            'type': StringFieldType,
            'description': 'The new branch name.',
        },
        'bugs_closed': {
            'type': StringFieldType,
            'description': 'A comma-separated list of bug IDs.',
        },
        'commit_id': {
            'type': StringFieldType,
            'description': 'The updated ID of the commit this review request '
                           'is based upon.',
            'added_in': '2.0',
        },
        'depends_on': {
            'type': StringFieldType,
            'description': 'The new list of dependencies of this review '
                           'request.',
            'added_in': '1.7.8',
        },
        'changedescription': {
            'type': StringFieldType,
            'description': 'The change description for this update.',
            'supports_text_types': True,
        },
        'changedescription_text_type': {
            'type': ChoiceFieldType,
            'choices': MarkdownFieldsMixin.SAVEABLE_TEXT_TYPES,
            'description': 'The text type used for the ``changedescription`` '
                           'field.',
            'added_in': '2.0.12',
        },
        'description': {
            'type': StringFieldType,
            'description': 'The new review request description.',
            'supports_text_types': True,
        },
        'description_text_type': {
            'type': ChoiceFieldType,
            'choices': MarkdownFieldsMixin.SAVEABLE_TEXT_TYPES,
            'description': 'The text type used for the ``description`` '
                           'field.',
            'added_in': '2.0.12',
        },
        'force_text_type': {
            'type': ChoiceFieldType,
            'choices': MarkdownFieldsMixin.TEXT_TYPES,
            'description': 'The text type, if any, to force for returned '
                           'text fields. The contents will be converted '
                           'to the requested type in the payload, but '
                           'will not be saved as that type.',
            'added_in': '2.0.9',
        },
        'public': {
            'type': BooleanFieldType,
            'description': 'Whether or not to make the review public. '
                           'If a review is public, it cannot be made '
                           'private again.',
        },
        'publish_as_owner': {
            'type': bool,
            'description': 'Publish on behalf of the owner of the review '
                           'request. If setting ``submitter``, this will '
                           'publish on behalf of the previous owner.',
            'added_in': '3.0.6',
        },
        'submitter': {
            'type': StringFieldType,
            'description': 'The user who submitted the review request.',
            'added_in': '3.0',
        },
        'summary': {
            'type': StringFieldType,
            'description': 'The new review request summary.',
        },
        'target_groups': {
            'type': StringFieldType,
            'description': 'A comma-separated list of review groups '
                           'that will be on the reviewer list.',
        },
        'target_people': {
            'type': StringFieldType,
            'description': 'A comma-separated list of users that will '
                           'be on a reviewer list.',
        },
        'testing_done': {
            'type': StringFieldType,
            'description': 'The new testing done text.',
            'supports_text_types': True,
        },
        'testing_done_text_type': {
            'type': ChoiceFieldType,
            'choices': MarkdownFieldsMixin.SAVEABLE_TEXT_TYPES,
            'description': 'The text type used for the ``testing_done`` '
                           'field.',
            'added_in': '2.0.12',
        },
        'text_type': {
            'type': ChoiceFieldType,
            'choices': MarkdownFieldsMixin.SAVEABLE_TEXT_TYPES,
            'description': 'The mode for the ``changedescription``, '
                           '``description``, and ``testing_done`` fields.\n'
                           '\n'
                           'This is deprecated. Please use '
                           '``changedescription_text_type``, '
                           '``description_text_type``, and '
                           '``testing_done_text_type`` instead.',
            'added_in': '2.0',
            'deprecated_in': '2.0.12',
        },
        'trivial': {
            'type': BooleanFieldType,
            'description': 'Determines if the review request publish '
                           'will not send an email.',
            'added_in': '2.5',
        },
        'update_from_commit_id': {
            'type': BooleanFieldType,
            'description': 'If true, and if ``commit_id`` is provided, '
                           'the review request information and (when '
                           'supported) the diff will be updated based '
                           'on the commit ID.',
            'added_in': '2.0',
        },
    }

    VALUE_LIST_RE = re.compile(r'[, ]+')

    @classmethod
    def prepare_draft(cls, request, review_request):
        """Create a draft, if the user has permission to."""
        if not review_request.is_mutable_by(request.user):
            raise PermissionDenied

        return ReviewRequestDraft.create(review_request)

    def get_queryset(
        self,
        request: HttpRequest,
        *args,
        **kwargs,
    ) -> QuerySet:
        """Return a queryset for the draft.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            *args (tuple):
                Positional arguments to pass through to the review request
                resource.

            **kwargs (dict):
                Keyword arguments to pass through to the review request
                resource.

        Returns:
            django.db.models.QuerySet:
            The queryset for the review request draft.
        """
        review_request = resources.review_request.get_object(
            request, *args, **kwargs)

        return self.model.objects.filter(review_request=review_request)

    def get_is_changedescription_rich_text(self, obj):
        return obj.changedesc_id is not None and obj.changedesc.rich_text

    def serialize_bugs_closed_field(self, obj, **kwargs):
        return obj.get_bug_list()

    def serialize_changedescription_field(self, obj, **kwargs):
        if obj.changedesc:
            return obj.changedesc.text
        else:
            return ''

    def serialize_changedescription_text_type_field(self, obj, **kwargs):
        # This will be overridden by MarkdownFieldsMixin.
        return None

    def serialize_description_text_type_field(self, obj, **kwargs):
        # This will be overridden by MarkdownFieldsMixin.
        return None

    def serialize_status_field(self, obj, **kwargs):
        return ReviewRequest.status_to_string(obj.status)

    def serialize_public_field(self, obj, **kwargs):
        return False

    def serialize_testing_done_text_type_field(self, obj, **kwargs):
        # This will be overridden by MarkdownFieldsMixin.
        return None

    def get_extra_data_field_supports_markdown(self, review_request, key):
        field_cls = get_review_request_field(key)

        return field_cls and getattr(field_cls, 'enable_markdown', False)

    def has_access_permissions(self, request, draft, *args, **kwargs):
        return draft.is_accessible_by(request.user)

    def has_modify_permissions(self, request, draft, *args, **kwargs):
        return draft.is_mutable_by(request.user)

    def has_delete_permissions(self, request, draft, *args, **kwargs):
        return draft.is_mutable_by(request.user)

    @webapi_check_local_site
    @webapi_login_required
    @webapi_response_errors(
        COMMIT_ID_ALREADY_EXISTS,
        DOES_NOT_EXIST,
        INVALID_CHANGE_NUMBER,
        INVALID_FORM_DATA,
        NOT_LOGGED_IN,
        PERMISSION_DENIED,
        PUBLISH_ERROR,
        UNVERIFIED_HOST_CERT,
    )
    @webapi_request_fields(
        optional=CREATE_UPDATE_OPTIONAL_FIELDS,
        allow_unknown=True
    )
    def create(self,
               request,
               local_site_name=None,
               *args,
               **kwargs):
        """Creates a draft of a review request.

        If a draft already exists, this will just reuse the existing draft.

        See the documentation on updating a draft for all the details.
        """
        try:
            review_request = resources.review_request.get_object(
                request, local_site_name=local_site_name, *args, **kwargs)
        except ReviewRequest.DoesNotExist:
            return DOES_NOT_EXIST

        if review_request.status == ReviewRequest.DISCARDED:
            review_request.reopen(request.user)

        # A draft is a singleton. Creating and updating it are the same
        # operations in practice.
        result = self.update(request, local_site_name=local_site_name,
                             *args, **kwargs)

        if isinstance(result, tuple):
            if result[0] == 200:
                return (201,) + result[1:]

        return result

    @webapi_check_local_site
    @webapi_login_required
    @webapi_response_errors(
        COMMIT_ID_ALREADY_EXISTS,
        DOES_NOT_EXIST,
        INVALID_CHANGE_NUMBER,
        INVALID_FORM_DATA,
        NOT_LOGGED_IN,
        PERMISSION_DENIED,
        PUBLISH_ERROR,
        REPO_INFO_ERROR,
        UNVERIFIED_HOST_CERT,
    )
    @webapi_request_fields(
        optional=CREATE_UPDATE_OPTIONAL_FIELDS,
        allow_unknown=True
    )
    def update(self,
               request,
               local_site_name=None,
               branch=None,
               bugs_closed=None,
               changedescription=None,
               commit_id=None,
               depends_on=None,
               submitter=None,
               summary=None,
               target_groups=None,
               target_people=None,
               update_from_commit_id=False,
               trivial=None,
               publish_as_owner=False,
               extra_fields={},
               *args,
               **kwargs):
        """Updates a draft of a review request.

        This will update the draft with the newly provided data.

        Most of the fields correspond to fields in the review request, but
        there is one special one, ``public``. When ``public`` is set to true,
        the draft will be published, moving the new content to the
        review request itself, making it public, and sending out a notification
        (such as an e-mail) if configured on the server. The current draft will
        then be deleted.

        Extra data can be stored later lookup. See
        :ref:`webapi2.0-extra-data` for more information.
        """
        try:
            review_request = resources.review_request.get_object(
                request, local_site_name=local_site_name, *args, **kwargs)
        except ReviewRequest.DoesNotExist:
            return DOES_NOT_EXIST

        # This explicitly does not pass the user in because we do access
        # checking in has_modify_permissions.
        draft = review_request.get_draft()

        if draft:
            if not self.has_modify_permissions(request, draft):
                return self.get_no_access_error(request)
        elif not review_request.is_mutable_by(request.user):
            return self.get_no_access_error(request)

        # Before we update anything, sanitize the commit ID, see if it
        # changed, and make sure that the the new value isn't owned by
        # another review request or draft.
        if commit_id == '':
            commit_id = None

        if (commit_id and
            commit_id != review_request.commit_id and
            (draft is None or commit_id != draft.commit_id)):
            # The commit ID has changed, so now we check for other usages of
            # this ID.
            repository = review_request.repository

            existing_review_request_ids = (
                ReviewRequest.objects
                .filter(commit_id=commit_id,
                        repository=repository)
                .values_list('pk', flat=True)
            )

            if (existing_review_request_ids and
                review_request.pk not in existing_review_request_ids):
                # Another review request is using this ID. Error out.
                return COMMIT_ID_ALREADY_EXISTS

            existing_draft_ids = (
                ReviewRequestDraft.objects
                .filter(commit_id=commit_id,
                        review_request__repository=repository)
                .values_list('pk', flat=True)
            )

            if (existing_draft_ids and
                (draft is None or draft.pk not in existing_draft_ids)):
                # Another review request draft is using this ID. Error out.
                return COMMIT_ID_ALREADY_EXISTS

        # Now that we've completed our initial accessibility and conflict
        # checks, we can start checking for changes to individual fields.
        #
        # We'll keep track of state pertaining to the fields we want to
        # set/save, and any errors we hit. For setting/saving, these's two
        # types of things we're tracking: The new field values (which will be
        # applied to the objects or Many-To-Many relations) and a list of
        # field names to set when calling `save(update_fields=...)`. The
        # former implies the latter. The latter only needs to be directly
        # set if the fields are modified directly by another function.
        new_draft_values = {}
        new_changedesc_values = {}
        draft_update_fields = set()
        changedesc_update_fields = set()
        invalid_fields = {}

        if draft is None:
            draft = ReviewRequestDraft.create(review_request=review_request)

        # Check for a new value for branch.
        if branch is not None:
            new_draft_values['branch'] = branch

        # Check for a new value for bugs_closed:
        if bugs_closed is not None:
            new_draft_values['bugs_closed'] = \
                ','.join(self._parse_bug_list(bugs_closed))

        # Check for a new value for changedescription.
        if changedescription is not None and draft.changedesc_id is None:
            invalid_fields['changedescription'] = [
                'Change descriptions cannot be used for drafts of '
                'new review requests'
            ]

        # Check for a new value for commit_id.
        if commit_id is not None:
            new_draft_values['commit_id'] = commit_id

            if update_from_commit_id:
                try:
                    draft_update_fields.update(
                        draft.update_from_commit_id(commit_id))
                except InvalidChangeNumberError:
                    return INVALID_CHANGE_NUMBER
                except CertificateVerificationError as e:
                    return UNVERIFIED_HOST_CERT.with_message(str(e)), {
                        'certificate': e.certificate,
                    }
                except (HostingServiceError, SCMError) as e:
                    return REPO_INFO_ERROR.with_message(str(e))

        # Check for a new value for depends_on.
        if depends_on is not None:
            found_deps, missing_dep_ids = self._find_depends_on(
                dep_ids=self._parse_value_list(depends_on),
                request=request)

            if missing_dep_ids:
                invalid_fields['depends_on'] = missing_dep_ids
            else:
                new_draft_values['depends_on'] = found_deps

        # Check for a new value for submitter.
        if submitter is not None:
            # While we only allow for one submitter, we'll try to parse a
            # possible list of values. This allows us to provide a suitable
            # error message if users try to set more than one submitter
            # (which people do try, in practice).
            found_users, missing_usernames = self._find_users(
                usernames=self._parse_value_list(submitter),
                request=request)

            if len(found_users) + len(missing_usernames) > 1:
                invalid_fields['submitter'] = [
                    'Only one user can be set as the owner of a review '
                    'request'
                ]
            elif missing_usernames:
                assert len(missing_usernames) == 1
                invalid_fields['submitter'] = missing_usernames
            elif found_users:
                assert len(found_users) == 1
                new_draft_values['owner'] = found_users[0]
            else:
                invalid_fields['submitter'] = [
                    'The owner of a review request cannot be blank'
                ]

        # Check for a new value for summary.
        if summary is not None:
            if '\n' in summary:
                invalid_fields['summary'] = [
                    "The summary can't contain a newline"
                ]
            else:
                new_draft_values['summary'] = summary

        # Check for a new value for target_groups.
        if target_groups is not None:
            found_groups, missing_group_names = self._find_review_groups(
                group_names=self._parse_value_list(target_groups),
                request=request)

            if missing_group_names:
                invalid_fields['target_groups'] = missing_group_names
            else:
                new_draft_values['target_groups'] = found_groups

        # Check for a new value for target_people.
        if target_people is not None:
            found_users, missing_usernames = self._find_users(
                usernames=self._parse_value_list(target_people),
                request=request)

            if missing_usernames:
                invalid_fields['target_people'] = missing_usernames
            else:
                new_draft_values['target_people'] = found_users

        # See if we've caught any invalid values. If so, we can error out
        # immediately before we update anything else.
        if invalid_fields:
            return INVALID_FORM_DATA, {
                'fields': invalid_fields,
                self.item_result_key: draft,
            }

        # Start applying any rich text processing to any text fields on the
        # ChangeDescription and draft. We'll track any fields that get set
        # for later saving.
        #
        # NOTE: If any text fields or text type fields are ever made
        #       parameters of this method, then their values will need to be
        #       passed directly to set_text_fields() calls below.
        if draft.changedesc_id:
            changedesc_update_fields.update(
                self.set_text_fields(obj=draft.changedesc,
                                     text_field='changedescription',
                                     text_model_field='text',
                                     rich_text_field_name='rich_text',
                                     changedescription=changedescription,
                                     **kwargs))

        for text_field in ('description', 'testing_done'):
            draft_update_fields.update(self.set_text_fields(
                obj=draft,
                text_field=text_field,
                **kwargs))

        # Go through the list of Markdown-enabled custom fields and apply
        # any rich text processing. These will go in extra_data, so we'll
        # want to make sure that's tracked for saving.
        for field_cls in get_review_request_fields():
            if (not issubclass(field_cls, BuiltinFieldMixin) and
                getattr(field_cls, 'enable_markdown', False)):
                modified_fields = self.set_extra_data_text_fields(
                    obj=draft,
                    text_field=field_cls.field_id,
                    extra_fields=extra_fields,
                    **kwargs)

                if modified_fields:
                    draft_update_fields.add('extra_data')

        # See if the caller has set or patched extra_data values. For
        # compatibility, we're going to do this after processing the rich
        # text fields in extra_data above.
        try:
            if self.import_extra_data(draft, draft.extra_data, extra_fields):
                # Track extra_data for saving.
                draft_update_fields.add('extra_data')
        except ImportExtraDataError as e:
            return e.error_payload

        # Everything checks out. We can now begin the process of applying any
        # field changes and then save objects.
        #
        # We'll start by making an initial pass on the objects we need to
        # either care about. This optimistically lets us avoid a lookup on the
        # ChangeDescription, if it's not being modified.
        to_apply = []

        if draft_update_fields or new_draft_values:
            # If there's any changes made at all to the draft, make sure we
            # allow last_updated to be computed and saved.
            draft_update_fields.add('last_updated')

            to_apply.append((draft, draft_update_fields, new_draft_values))

        if changedesc_update_fields or new_changedesc_values:
            to_apply.append((draft.changedesc, changedesc_update_fields,
                             new_changedesc_values))

        for obj, update_fields, new_values in to_apply:
            new_m2m_values = {}

            # We may have a mixture of field values and Many-To-Many
            # relation values, which we want to set only after the object
            # is saved. Start by setting any field values, and store the
            # M2M values for after.
            for key, value in new_values.items():
                field = obj._meta.get_field(key)

                if isinstance(field, ManyToManyField):
                    # Save this until after the object is saved.
                    new_m2m_values[key] = value
                else:
                    # We can set this one now, and mark it for saving.
                    setattr(obj, key, value)
                    update_fields.add(key)

            if update_fields:
                obj.save(update_fields=sorted(update_fields))

            # Now we can set any values on M2M fields.
            #
            # Each entry will have zero or more values. We'll be
            # setting to the list of values, which will fully replace
            # the stored entries in the database.
            for key, values in new_m2m_values.items():
                field = getattr(obj, key)
                field.set(values)

        # Next, check if the draft is set to be published.
        if request.POST.get('public', False):
            if not review_request.public and not draft.changedesc_id:
                # This is a new review request. Publish this on behalf of the
                # owner of the review request, rather than the current user,
                # regardless of the original publish_as_owner in the request.
                # This allows a review request previously created with
                # submit-as= to be published by that user instead of the
                # logged in user.
                publish_as_owner = True

            if publish_as_owner:
                publish_user = review_request.owner
            else:
                # Default to posting as the logged in user.
                publish_user = request.user

            try:
                review_request.publish(user=publish_user, trivial=trivial)
            except NotModifiedError:
                return NOTHING_TO_PUBLISH
            except PublishError as e:
                return PUBLISH_ERROR.with_message(str(e))

        return 200, {
            self.item_result_key: draft,
        }

    @webapi_check_local_site
    @webapi_login_required
    @webapi_response_errors(DOES_NOT_EXIST, NOT_LOGGED_IN, PERMISSION_DENIED)
    def delete(self, request, *args, **kwargs):
        """Deletes a draft of a review request.

        This is equivalent to pressing :guilabel:`Discard Draft` in the
        review request's page. It will simply erase all the contents of
        the draft.
        """
        # Make sure this exists.
        try:
            review_request = \
                resources.review_request.get_object(request, *args, **kwargs)
        except ReviewRequest.DoesNotExist:
            return DOES_NOT_EXIST

        # This explicitly does not pass the user in because we delete
        # checking in has_modify_permissions.
        draft = review_request.get_draft()

        if draft is None:
            return DOES_NOT_EXIST

        if not self.has_delete_permissions(request, draft, *args, **kwargs):
            return self.get_no_access_error(request)

        draft.delete()

        return 204, {}

    @webapi_login_required
    @augment_method_from(WebAPIResource)
    def get(self, request, review_request_id, *args, **kwargs):
        """Returns the current draft of a review request."""
        pass

    def get_links(self, child_resources=[], obj=None, request=None,
                  *args, **kwargs):
        """Return the links for the resource.

        This method will filter out the draft diffcommit resource when the DVCS
        feature is disabled.

        Args:
            child_resources (list of djblets.webapi.resources.base.
                             WebAPIResource):
                The child resources for which links will be serialized.

            review_request_id (unicode):
                A string representation of the ID of the review request for
                which links are being returned.

            request (django.http.HttpRequest):
                The HTTP request from the client.

            *args (tuple):
                Additional positional arguments.

            **kwargs (dict):
                Additional keyword arguments.

        Returns:
            dict:
            A dictionary of the links for the resource.
        """
        if (obj is not None and
            not dvcs_feature.is_enabled() and
            resources.draft_diffcommit in child_resources):
            child_resources = list(child_resources)
            child_resources.remove(resources.draft_diffcommit)

        return super(ReviewRequestDraftResource, self).get_links(
            child_resources, obj=obj, request=request, *args, **kwargs)

    def _parse_value_list(self, data):
        """Parse a list of values from a string.

        This will parse a comma-separated list of values into a list of
        strings. All items will be stripped, and any empty values will be
        removed.

        Args:
            data (unicode):
                The data to parse.

        Returns:
            list of unicode:
            The parsed list of strings.
        """
        return [
            value
            for value in [
                value.strip()
                for value in self.VALUE_LIST_RE.split(data)
            ]
            if value
        ]

    def _parse_bug_list(self, bug_ids):
        """Parse a list of bug IDs.

        This will remove any excess whitespace before or after the bug
        IDs, and remove any leading ``#`` characters.

        Args:
            bug_ids (unicode):
                The comma-separated list of bug IDs to parse.

        Returns:
            list of unicode:
            The parsed list of bug IDs.
        """
        return [
            # RB stores bug numbers as numbers, but many people have the
            # habit of prepending #, so filter it out:
            bug.lstrip('#')
            for bug in self._parse_value_list(bug_ids)
        ]

    def _find_depends_on(self, dep_ids, request):
        """Return any found and missing review request dependencies.

        This will look up :py:class:`ReviewRequests
        <reviewboard.reviews.models.review_request.ReviewRequest>` that are
        dependencies with the given list of IDs.

        Args:
            dep_ids (list of unicode):
                The list of review request IDs to look up.

            request (django.http.HttpRequest):
                The HTTP request from the client.

        Return:
            tuple:
            A tuple containing:

            1. A list of :py:class:`~reviewboard.reviews.models.
               review_request.ReviewRequest` instances for any IDs that
               were found.

            2. A list of IDs that could not be found.
        """
        if not dep_ids:
            return [], []

        query_ids = []

        # Filter out anything that isn't an integer.
        for dep_id in dep_ids:
            try:
                query_ids.append(int(dep_id))
            except ValueError:
                pass

        local_site = request.local_site

        if local_site:
            review_requests = ReviewRequest.objects.filter(
                local_site=local_site,
                local_id__in=query_ids)
        else:
            review_requests = ReviewRequest.objects.filter(pk__in=query_ids)

        review_requests = list(review_requests)
        missing_dep_ids = []

        if len(review_requests) != len(dep_ids):
            # Some review requests couldn't be found. Find out which.
            found_dep_ids = set(
                review_request.display_id
                for review_request in review_requests
            )

            missing_dep_ids = [
                dep_id
                for dep_id in dep_ids
                if dep_id not in found_dep_ids
            ]

        return review_requests, missing_dep_ids

    def _find_review_groups(self, group_names, request):
        """Return any found and missing review groups based on a list of names.

        This will look up :py:class:`Groups
        <reviewboard.site.models.group.Group>` from the database based on the
        list of group names.

        Args:
            group_names (list of unicode):
                The list of group names to look up.

            request (django.http.HttpRequest):
                The HTTP request from the client.

        Return:
            tuple:
            A tuple containing:

            1. A list of :py:class:`~reviewboard.site.models.group.Group`
               instances for any group names that were found.

            2. A list of group names that could not be found.
        """
        if not group_names:
            return [], []

        # Build a query that will find each group with a case-insensitive
        # search.
        q = Q()

        for group_name in group_names:
            q |= Q(name__iexact=group_name)

        groups = (
            Group.objects
            .filter(local_site=request.local_site)
            .filter(q)
        )
        missing_group_names = []

        if len(group_names) != len(groups):
            # Some groups couldn't be found. Find out which.
            found_group_names = set(
                group.name
                for group in groups
            )

            missing_group_names = [
                group_name
                for group_name in group_names
                if group_name not in found_group_names
            ]

        return groups, missing_group_names

    def _find_users(self, usernames, request):
        """Return any found and missing users based on a list of usernames.

        This will look up :py:meth:`Users <django.contrib.auth.models.User>`
        from the database based on the list of usernames. If the request is
        not performed on a :term:`Local Site`, then this will then attempt
        to look up any missing users from the authentication backends,
        creating them as necessary.

        Args:
            usernames (list of unicode):
                The list of usernames to look up.

            request (django.http.HttpRequest):
                The HTTP request from the client.

        Return:
            tuple:
            A tuple containing:

            1. A list of :py:class:`~django.contrib.auth.models.User`
               instances for any usernames that were found.

            2. A list of usernames that could not be found.
        """
        if not usernames:
            return [], []

        local_site = request.local_site

        if local_site:
            users = local_site.users.filter(username__in=usernames)
        else:
            users = User.objects.filter(username__in=usernames)

        users = list(users)
        missing_usernames = []

        if len(users) != len(usernames):
            # Some users couldn't be found. Find out which.
            found_usernames = set(
                user.username
                for user in users
            )

            if not local_site:
                # See if any of these users exist in an auth backend.
                # We don't do this for Local Sites, since we don't want to
                # risk creating users in sites where they don't belong.
                for username in usernames:
                    if username in found_usernames:
                        continue

                    for backend in auth.get_backends():
                        try:
                            user = backend.get_or_create_user(username,
                                                              request)

                            if user is not None:
                                users.append(user)
                                found_usernames.add(username)
                        except NotImplementedError:
                            pass
                        except Exception as e:
                            logger.exception(
                                'Error when calling get_or_create_user for '
                                'auth backend %r: %s',
                                backend, e)

            if len(users) != len(usernames):
                missing_usernames = [
                    username
                    for username in usernames
                    if username not in found_usernames
                ]

        return users, missing_usernames


review_request_draft_resource = ReviewRequestDraftResource()



"""Add Profile.open_an_issue.

Version Added:
    1.7
"""

from django_evolution.mutations import AddField
from django.db import models


MUTATIONS = [
    AddField('Profile', 'open_an_issue', models.BooleanField, initial=True)
]



"""Base support for code safety checkers.

Version Added:
    5.0
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from django.template.loader import render_to_string
from django.utils.safestring import SafeString, mark_safe
from typing_extensions import NotRequired, TypedDict

if TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import Any

    from typelets.django.strings import StrOrPromise

    from reviewboard.scmtools.models import Repository


class CodeSafetyContentItem(TypedDict):
    """An item of content in a file to check.

    Version Added:
        5.0.2
    """

    #: The path to the file in a diff to check.
    #:
    #: This can be used by checkers to perform checks on only certain kinds
    #: of files, or to change checking behavior based on the file type.
    #:
    #: This can just be a filename, rather than a full path, if that's all
    #: that's available.
    #:
    #: Type:
    #:     str
    path: str

    #: A list of one or more lines within the file to check.
    #:
    #: The checker cannot assume anything about the range of lines within the
    #: file.
    #:
    #: Type:
    #:     list of str
    lines: list[str]

    #: The repository the file is on, if any.
    #:
    #: Type:
    #:     rbtools.scmtools.models.Repository
    repository: NotRequired[Repository]


class CodeSafetyCheckResults(TypedDict):
    """The results of a code safety check.

    Version Added:
        5.0.2
    """

    #: A set of error IDs found by a code safety checker.
    #:
    #: The IDs are local to the code safety checker.
    #:
    #: Type:
    #:     list of str
    errors: NotRequired[set[str]]

    #: A set of warning IDs found by a code safety checker.
    #:
    #: The IDs are local to the code safety checker.
    #:
    #: Type:
    #:     list of str
    warnings: NotRequired[set[str]]


class BaseCodeSafetyChecker:
    """Base class for a code safety checker.

    Code safety checkers are used to analyze the content of lines of code in
    order to determine if there's anything suspicious or dangerous.

    They're designed to be very general-purpose, allowing the caller to check
    a group of lines (potentially from different files or revisions of files)
    at once, in order to get a list of any results (warnings or errors) that
    were found.

    They can also update the rendering of lines to highlight those results,
    and show a detailed message designed to be shown at the top of a file or
    in an API result.

    Version Added:
        5.0
    """

    #: The unique ID of the code safety checker, for registration purposes.
    #:
    #: Type:
    #:     str
    checker_id: (str | None) = None

    #: The summary shown by the code safety checker for results.
    #:
    #: Type:
    #:    str
    summary: (StrOrPromise | None) = None

    #: The HTML template name for the alert at the top of a file.
    #:
    #: If provided, this will be used as the default template when rendering
    #: file alerts in :py:meth:`render_file_alert`.
    #:
    #: Type:
    #:     str
    file_alert_html_template_name: (str | None) = None

    #: A mapping of warning IDs to human-readable labels.
    #:
    #: These will be used whenever the details on a warning needs to be shown
    #: to a user.
    #:
    #: Keys must be strings, and values should be lazily-localized strings.
    #:
    #: Type:
    #:     dict
    result_labels: dict[str, StrOrPromise] = {}

    def check_content(
        self,
        content_items: list[CodeSafetyContentItem],
        **kwargs,
    ) -> CodeSafetyCheckResults:
        """Check content for safety issues.

        One or more files may be checked at once, and one or more lines
        somewhere within those files. The checker should return any warnings
        or errors that are found within any of those provided lines.

        It's up to the caller to decide whether a full file's content,
        multiple file's contents, or specific ranges from multiple files are
        checked, and to handle the results appropriately.

        Subclasses can extend this with custom arguments. These should all
        be specified as keyword-only arguments. Review Board may set these
        based on stored configuration, depending on the code safety checker.

        Version Changed:
            5.0.2:
            Added explicit support for subclass-defined custom arguments.

        Args:
            content_items (list of dict):
                A list of dictionaries containing files and lines to check.

                See :py:class:`CodeSafetyContentItem` for the contents of
                each item.

            **kwargs (dict, unused):
                Additional keyword arguments, for future expansion.

        Returns:
            dict:
            Results from the checks. See :py:class:`CodeSafetyCheckResults`
            for details.
        """
        return {}

    def update_line_html(
        self,
        line_html: str,
        result_ids: Sequence[str],
        **kwargs,
    ) -> SafeString:
        """Update the rendered diff HTML for a line.

        This can update the HTML for a line to highlight any content that
        triggered a warning, making the problem clear to reviewers or authors
        of changes.

        Callers should take care to ensure that the updates don't themselves
        cause any unsafe HTML to be generated.

        Subclasses can extend this with custom arguments. These should all
        be specified as keyword-only arguments. Review Board may set these
        based on stored configuration, depending on the code safety checker.

        Version Changed:
            5.0.2:
            Added explicit support for subclass-defined custom arguments.

        Args:
            line_html (str):
                The HTML of the line.

            result_ids (list of str):
                The list of result IDs that were found for the line.

            **kwargs (dict, unused):
                Additional keyword arguments, for future expansion.

        Returns:
            django.utils.safestring.SafeString:
            The updated HTML.
        """
        return mark_safe(line_html)

    def get_result_labels(
        self,
        result_ids: Sequence[str],
        **kwargs,
    ) -> list[str]:
        """Return a list of result labels for the given IDs for display.

        By default, this will generate a list based off
        :py:attr:`result_labels`. Subclasses can override this if they need
        to do something more specific.

        Args:
            result_ids (list of str):
                A list of one or more result IDs generated by this code
                safety checker.

            **kwargs (dict, unused):
                Additional keyword arguments, for future expansion.

        Returns:
            list of str:
            A list of result labels for the given IDs.
        """
        labels = self.result_labels

        # Note that we need to cast these to a string in case the text is
        # lazily-localized.
        return [
            str(labels.get(_result_id, _result_id))
            for _result_id in result_ids
        ]

    def render_file_alert_html(
        self,
        error_ids: Sequence[str],
        warning_ids: Sequence[str],
        **kwargs,
    ) -> SafeString | None:
        """Render an alert for the top of a file.

        This is responsible for rendering an alert that explains the warnings
        and errors that were found. This would be shown at the top of a diff.

        By default, this will render :py:attr:`file_alert_html_template_name`,
        if specified, and provide context variables from
        :py:meth:`get_file_alert_context_data`.

        Subclasses can override this to provide additional information.

        Args:
            error_ids (list of str):
                A list of one or more error IDs generated by this code
                safety checker.

            warning_ids (list of str):
                A list of one or more warning IDs generated by this code
                safety checker.

            **kwargs (dict, unused):
                Additional keyword arguments, for future expansion.

        Returns:
            django.utils.safestring.SafeString:
            The alert HTML, or ``None`` if nothing will be displayed.
        """
        if not self.file_alert_html_template_name:
            return None

        context_data = self.get_file_alert_context_data(
            error_ids=error_ids or [],
            warning_ids=warning_ids or [],
            **kwargs)

        return render_to_string(self.file_alert_html_template_name,
                                context=context_data)

    def get_file_alert_context_data(
        self,
        error_ids: Sequence[str],
        warning_ids: Sequence[str],
        **kwargs,
    ) -> dict[str, Any]:
        """Return context variables for the file alert template.

        By default, this returns:

        Keys:
            error_ids (list of str):
                The list of error IDs found by this code safety checker.

            error_labels (list of str):
                A list of human-readable warning labels for display.

            warning_ids (list of str):
                The list of warning IDs found by this code safety checker.

            warning_labels (list of str):
                A list of human-readable warning labels for display.

        Subclasses can override this to return additional data.

        Args:
            error_ids (list of str):
                A list of one or more error IDs generated by this code
                safety checker.

            warning_ids (list of str):
                A list of one or more warning IDs generated by this code
                safety checker.

            **kwargs (dict, unused):
                Additional keyword arguments, for future expansion.

        Returns:
            dict:
            A dictionary of context variables.
        """
        return {
            'error_ids': error_ids,
            'error_labels': self.get_result_labels(error_ids),
            'warning_ids': warning_ids,
            'warning_labels': self.get_result_labels(warning_ids),
        }



"""Unit tests for reviewboard.accounts.forms.pages.AccountSettingsForm."""

from django.contrib.auth.models import User
from django.views.generic.base import View
from djblets.testing.decorators import add_fixtures

from reviewboard.accounts.forms.pages import AccountSettingsForm
from reviewboard.accounts.models import Profile
from reviewboard.accounts.pages import AccountSettingsPage
from reviewboard.testing import TestCase


class AccountSettingsFormTests(TestCase):
    """Unit tests for reviewboard.accounts.forms.pages.AccountSettingsForm."""

    @add_fixtures(['test_users'])
    def test_save_syntax_highlighting_disabled(self):
        """Testing AccountSettingsForm.save() with
        diffviewer_syntax_highlighting disabled
        """
        view = View()
        user = User.objects.get(username='doc')

        profile = user.get_profile()
        profile.syntax_highlighting = True
        profile.save(update_fields=('syntax_highlighting',))

        request = self.create_http_request()
        page = AccountSettingsPage(view, request, user)

        settings = {'diffviewer_syntax_highlighting': False}

        with self.siteconfig_settings(settings):
            form = AccountSettingsForm(page, request, user, data={
                'syntax_highlighting': False,
                'timezone': profile.timezone,
            })

            self.assertTrue(form.is_valid())
            form.save()

        profile = Profile.objects.get(pk=profile.pk)
        self.assertTrue(profile.syntax_highlighting)



"""Errors for hosting service operations."""

from __future__ import annotations

from django.utils.translation import gettext, gettext_lazy as _


class HostingServiceError(Exception):
    """Base class for errors related to a hosting service."""

    ######################
    # Instance variables #
    ######################

    #: The optional HTTP status code for the error.
    http_code: int | None

    #: An optional URL for further information on the error.
    help_link: str | None

    #: A label for the optional URL for further information on the error.
    #:
    #: This will always be provided if :py:attr:`help_link` is set.
    help_link_text: str | None

    def __init__(
        self,
        message: str,
        http_code: (int | None) = None,
        help_link: (str | None) = None,
        help_link_text: (str | None) = None,
    ) -> None:
        """Initialize the error.

        Args:
            message (str):
                The error message.

            http_code (int, optional):
                An optional HTTP status code for the error.

            help_link (str, optional):
                An optional URL for further information on the error.

            help_link_text (str, optional):
                A label for the optional URL for further information on the
                error.

                This must be provided if ``help_link`` is set.

        Raises:
            ValueError:
                ``help_link`` was provided but ``help_link_text`` was not.
        """
        super().__init__(message)

        self.http_code = http_code
        self.help_link = help_link
        self.help_link_text = help_link_text

        if help_link and not help_link_text:
            raise ValueError('help_link_text cannot be empty')


class HostingServiceAPIError(HostingServiceError):
    """An API error from a hosting service.

    This represents an error from the hosting service's API. It contains
    the error message, HTTP code, and the parsed response payload.

    HostingServiceAPIError instances can be further turned into other error
    types, or raised directly, if the error message is suitable for display.
    """

    ######################
    # Instance variables #
    ######################

    #: The HTTP status code for the error, if available.
    http_code: int | None

    #: The parsed payload for the error, if available.
    rsp: object | None

    def __init__(
        self,
        msg: (str | None) = None,
        http_code: (int | None) = None,
        rsp: (object | None) = None,
    ) -> None:
        """Initialize the error.

        Args:
            msg (str, optional):
                The error message from the service.

            http_code (int, optional):
                The HTTP code for the error.

            rsp (object, optional):
                The parsed payload for the error.
        """
        super().__init__(msg or '')

        self.http_code = http_code
        self.rsp = rsp


class RepositoryError(HostingServiceError):
    """An error validating, configuring or using a repository."""


class AuthorizationError(HostingServiceError):
    """An error authorizing an account with a hosting service."""


class MissingHostingServiceError(HostingServiceError):
    """Indicates that the hosting service could not be loaded."""

    ######################
    # Instance variables #
    ######################

    #: The ID of the hosting service associated with this error.
    #:
    #: Type:
    #:     str
    hosting_service_id: str

    def __init__(
        self,
        hosting_service_id: str,
        repository: (str | None) = None,
    ) -> None:
        """Initialize the error.

        Args:
            hosting_service_id (str):
                The ID of the hosting service that cannot be loaded.

            repository (str, optional):
                The name of the repository that is loading the hosting service.
        """
        if repository is not None:
            message = _(
                'The repository "%(repository)s" cannot load the hosting '
                'service "%(hosting_service)s". An administrator should '
                'ensure all necessary packages and extensions are installed.'
            ) % {
                'repository': repository,
                'hosting_service': hosting_service_id,
            }
        else:
            message = _(
                'The hosting service "%s" could not be loaded. An '
                'administrator should ensure all necessary packages and '
                'extensions are installed.'
            ) % hosting_service_id

        super().__init__(message)

        self.hosting_service_id = hosting_service_id


class TwoFactorAuthCodeRequiredError(AuthorizationError):
    """Response from a service indicating a two-factor auth code is required.

    Some services will indicate, after an authorization attempt, that a
    two-factor authorization code must be provided to complete authorization.
    Services can raise this error, along with a helpful message, to
    inform the user and the repository form of this.
    """


class InvalidPlanError(HostingServiceError):
    """Indicates an invalid plan name was used."""

    def __init__(
        self,
        plan: str,
    ) -> None:
        """Initialize the error.

        Args:
            plan (str):
                The ID of the plan.
        """
        super().__init__(
            gettext('%s is not a valid plan for this hosting service')
            % plan)


class SSHKeyAssociationError(HostingServiceError):
    """An error associating an SSH key with an account on a hosting service."""



"""Unit tests for reviewboard.admin.forms.SearchSettingsForm."""

import shutil
import tempfile

from django.forms import ValidationError
from djblets.siteconfig.models import SiteConfiguration

from reviewboard.admin.forms.search_settings import SearchSettingsForm
from reviewboard.search import search_backend_registry
from reviewboard.search.search_backends.base import (SearchBackend,
                                                     SearchBackendForm)
from reviewboard.search.search_backends.whoosh import WhooshBackend
from reviewboard.testing.testcase import TestCase


class SearchSettingsFormTests(TestCase):
    """Unit tests for reviewboard.admin.forms.SearchSettingsForm."""

    def test_init(self):
        """Testing SearchSettingsForm.__init__"""
        siteconfig = SiteConfiguration.objects.get_current()
        form = SearchSettingsForm(siteconfig)
        subforms = form.search_backend_forms

        self.assertFalse(subforms['elasticsearch'].is_bound)
        self.assertFalse(subforms['whoosh'].is_bound)

    def test_init_with_data(self):
        """Testing SearchSettingsForm.__init__ with POST data"""
        siteconfig = SiteConfiguration.objects.get_current()
        form = SearchSettingsForm(siteconfig, data={
            'search_backend_id': 'whoosh',
        })
        subforms = form.search_backend_forms

        self.assertFalse(subforms['elasticsearch'].is_bound)
        self.assertTrue(subforms['whoosh'].is_bound)

    def test_clean(self):
        """Testing SearchSettingsForm.clean"""
        index_dir = tempfile.mkdtemp()

        try:
            siteconfig = SiteConfiguration.objects.get_current()
            form = SearchSettingsForm(siteconfig, data={
                'search_enable': True,
                'search_backend_id': WhooshBackend.search_backend_id,
                'whoosh-search_index_file': index_dir,
            })

            self.assertTrue(form.is_valid())
        finally:
            shutil.rmtree(index_dir)

    def test_clean_with_backend_validate_fail(self):
        """Testing SearchSettingsForm.clean when the backend doesn't pass
        validation
        """
        class InvalidSearchBackendForm(SearchBackendForm):
            pass

        class InvalidSearchBackend(SearchBackend):
            search_backend_id = 'invalid'
            config_form_class = InvalidSearchBackendForm

            def validate(self, **kwargs):
                raise ValidationError('This backend is invalid.')

        backend = InvalidSearchBackend()
        search_backend_registry.register(backend)

        try:
            siteconfig = SiteConfiguration.objects.get_current()
            form = SearchSettingsForm(siteconfig, data={
                'search_enable': True,
                'search_backend_id': backend.search_backend_id,
            })

            self.assertFalse(form.is_valid())
            self.assertEqual(form.errors['search_backend_id'],
                             ['This backend is invalid.'])
        finally:
            search_backend_registry.unregister(backend)

    def test_clean_missing_backend(self):
        """Testing SearchSettingsForm.clean when the backend doesn't exist"""
        siteconfig = SiteConfiguration.objects.get_current()
        form = SearchSettingsForm(siteconfig, data={
            'search_enable': True,
            'search_backend_id': 'non-existent',
        })

        self.assertFalse(form.is_valid())
        self.assertIn('search_backend_id', form.errors)



"""Unit tests for reviewboard.scmtools.core.SCMClient."""

from urllib.error import HTTPError

import kgb

from reviewboard.certs.http import urlopen
from reviewboard.scmtools.core import SCMClient
from reviewboard.scmtools.errors import FileNotFoundError, SCMError
from reviewboard.testing.testcase import TestCase


class GetFileHTTPResponse(object):
    def info(self):
        return {
            'Content-Type': 'text/plain',
        }

    def read(self):
        return b'abc'


class SCMClientTests(kgb.SpyAgency, TestCase):
    """Tests for reviewboard.scmtools.core.SCMClient."""

    def test_get_file_http(self):
        """Testing SCMClient.get_file_http"""
        self.spy_on(urlopen, op=kgb.SpyOpReturn(GetFileHTTPResponse()))

        client = SCMClient(path='/path/to/repo')

        self.assertEqual(client.get_file_http('https://example.com',
                                              path='/path/to/file',
                                              revision='abc123'),
                         b'abc')

    def test_get_file_http_with_username(self):
        """Testing SCMClient.get_file_http with username"""
        self.spy_on(urlopen, op=kgb.SpyOpReturn(GetFileHTTPResponse()))

        client = SCMClient(path='/path/to/repo',
                           username='test-user',
                           password='test-pass')

        self.assertEqual(client.get_file_http('https://example.com',
                                              path='/path/to/file',
                                              revision='abc123'),
                         b'abc')

        request = urlopen.last_call.args[0]
        self.assertEqual(request.headers[str('Authorization')],
                         str('Basic dGVzdC11c2VyOnRlc3QtcGFzcw=='))

    def test_get_file_http_with_mime_type_match(self):
        """Testing SCMClient.get_file_http with mime_type and match"""
        self.spy_on(urlopen, op=kgb.SpyOpReturn(GetFileHTTPResponse()))

        client = SCMClient(path='/path/to/repo')

        self.assertEqual(client.get_file_http('https://example.com',
                                              path='/path/to/file',
                                              revision='abc123',
                                              mime_type='text/plain'),
                         b'abc')

    def test_get_file_http_with_mime_type_no_match(self):
        """Testing SCMClient.get_file_http with mime_type and no match"""
        self.spy_on(urlopen, op=kgb.SpyOpReturn(GetFileHTTPResponse()))

        client = SCMClient(path='/path/to/repo')

        self.assertIsNone(client.get_file_http('https://example.com',
                                               path='/path/to/file',
                                               revision='abc123',
                                               mime_type='text/xxx'))

    def test_get_file_http_with_http_error(self):
        """Testing SCMClient.get_file_http with HTTPError"""
        self.spy_on(urlopen,
                    op=kgb.SpyOpRaise(HTTPError(url='https://example.com',
                                                code=500,
                                                msg='Kablam',
                                                hdrs=None,
                                                fp=None)))

        client = SCMClient(path='/path/to/repo')

        message = (
            r'Failed to fetch file at https://example.com. The repository '
            r'returned HTTP error 500\. Administrators can find details in '
            r'the Review Board server logs \(error ID [0-9a-z-]+\).'
        )

        with self.assertRaisesRegex(SCMError, message) as ctx:
            client.get_file_http('https://example.com',
                                 path='/path/to/file',
                                 revision='abc123')

        self.assertNotIsInstance(ctx.exception, FileNotFoundError)

    def test_get_file_http_with_http_error_404(self):
        """Testing SCMClient.get_file_http with HTTPError 404"""
        self.spy_on(urlopen,
                    op=kgb.SpyOpRaise(HTTPError(url='https://example.com',
                                                code=404,
                                                msg=None,
                                                hdrs=None,
                                                fp=None)))

        client = SCMClient(path='/path/to/repo')

        with self.assertRaises(FileNotFoundError) as ctx:
            client.get_file_http('https://example.com',
                                 path='/path/to/file',
                                 revision='abc123')

        e = ctx.exception
        self.assertEqual(e.path, '/path/to/file')
        self.assertEqual(e.revision, 'abc123')



"""Change ChangeDescription.fields_changed to a longtext type on MySQL.

Version Added:
    1.5.6
"""

from django_evolution.mutations import SQLMutation


MUTATIONS = [
    SQLMutation('mysql_fields_changed_longtext', ["""
        ALTER TABLE changedescs_changedescription
             MODIFY fields_changed LONGTEXT;
"""])
]



"""Base support for Review Board avatars."""

from djblets.registries.importer import lazy_import_registry


#: The avatar services registry.
avatar_services = lazy_import_registry('reviewboard.avatars.registry',
                                       'AvatarServiceRegistry')



"""Error definitions for attachments.

Version Added:
    5.0
"""

from __future__ import annotations


class FileTooBigError(ValueError):
    """The supplied file was too large.

    Version Added:
        5.0
    """

    ######################
    # Instance variables #
    ######################

    #: The maximum size (in bytes) of file attachments.
    max_attachment_size: int

    def __init__(
        self,
        message: str,
        *,
        max_attachment_size: int,
    ) -> None:
        """Initialize the error.

        Args:
            message (str):
                The error message to display.

            max_attachment_size (int):
                The maximum allowable attachment file size, in bytes.
        """
        super().__init__(message)
        self.max_attachment_size = max_attachment_size



"""Root resource for the Review Board API."""

from djblets.util.decorators import augment_method_from
from djblets.webapi.resources.root import RootResource as DjbletsRootResource

from reviewboard.webapi.server_info import get_server_info
from reviewboard.webapi.decorators import (webapi_check_login_required,
                                           webapi_check_local_site)
from reviewboard.webapi.resources import WebAPIResource, resources


class RootResource(WebAPIResource, DjbletsRootResource):
    """The root of the Review Board API resource tree.

    This should be used as a starting point for any clients that need to access
    any resources in the API. By browsing through the resource tree instead of
    hard-coding paths, your client can remain compatible with any changes in
    the resource URI scheme.

    This also contains information on the server and the capabilities of
    the API. This information was formerly provided only by the Server Info
    resource, but has been moved here as a convenience to clients.


    .. rubric:: URI Templates

    The following URI templates (found in the ``uri_templates`` key) can be
    used to help quickly reach the right API resource.

    .. note::

       This list may have changed between versions of Review Board, with some
       items added and some pointing to corrected URLs. In particular, due to
       a bug on some Python versions, there were differences in Review Board
       4.0.x and 5.0/5.0.1.

       Please check the specific URI templates for your version of Review Board
       if you are having any trouble.

    .. list-table::
       :header-rows: 1

       * - URI template key
         - Resource
         - Version Added

       * - ``all_diff_comments``
         - :ref:`webapi2.0-root-diff-comment-list-resource`
         - 5.0.2

       * - ``all_file_attachment_comments``
         - :ref:`webapi2.0-root-file-attachment-comment-list-resource`
         - 5.0.2

       * - ``all_general_comments``
         - :ref:`webapi2.0-root-general-comment-list-resource`
         - 5.0.1

       * - ``all_reviews``
         - :ref:`webapi2.0-root-review-list-resource`
         - 5.0.1

       * - ``api_token``
         - :ref:`webapi2.0-api-token-resource`
         -

       * - ``api_tokens``
         - :ref:`webapi2.0-api-token-list-resource`
         -

       * - ``archived_review_request``
         - :ref:`webapi2.0-archived-review-request-resource`
         -

       * - ``archived_review_requests``
         - :ref:`webapi2.0-archived-review-request-list-resource`
         -

       * - ``commit_validation``
         - :ref:`webapi2.0-validate-diff-commit-resource`
         -

       * - ``default_reviewer``
         - :ref:`webapi2.0-default-reviewer-resource`
         -

       * - ``default_reviewers``
         - :ref:`webapi2.0-default-reviewer-list-resource`
         -

       * - ``diff``
         - :ref:`webapi2.0-diff-resource`
         -

       * - ``diffs``
         - :ref:`webapi2.0-diff-list-resource`
         -

       * - ``diff_commit``
         - :ref:`webapi2.0-diff-commit-resource`
         - 5.0.2

       * - ``diff_commits``
         - :ref:`webapi2.0-diff-commit-list-resource`
         - 5.0.2

       * - ``diff_context``
         - :ref:`webapi2.0-diff-context-resource`
         -

       * - ``diff_file_attachment``
         - :ref:`webapi2.0-diff-file-attachment-resource`
         -

       * - ``diff_file_attachments``
         - :ref:`webapi2.0-diff-file-attachment-list-resource`
         -

       * - ``diff_validation``
         - :ref:`webapi2.0-validate-diff-resource`
         -

       * - ``extension``
         - :ref:`webapi2.0-extension-resource`
         -

       * - ``extensions``
         - :ref:`webapi2.0-extension-list-resource`
         -

       * - ``file_diff``
         - :ref:`webapi2.0-file-diff-resource`
         - 5.0.2

       * - ``file_diffs``
         - :ref:`webapi2.0-file-diff-list-resource`
         - 5.0.2

       * - ``file_diff_comments``
         - :ref:`webapi2.0-file-diff-comment-list-resource`
         -

       * - ``file_diff_original``
         - :ref:`webapi2.0-original-file-resource`
         - 5.0.2

       * - ``file_diff_patched``
         - :ref:`webapi2.0-patched-file-resource`
         - 5.0.2

       * - ``group``
         - :ref:`webapi2.0-review-group-resource`
         -

       * - ``groups``
         - :ref:`webapi2.0-review-group-list-resource`
         -

       * - ``hosting_service``
         - :ref:`webapi2.0-hosting-service-resource`
         -

       * - ``hosting_services``
         - :ref:`webapi2.0-hosting-service-list-resource`
         -

       * - ``hosting_service_account``
         - :ref:`webapi2.0-hosting-service-account-resource`
         -

       * - ``hosting_service_accounts``
         - :ref:`webapi2.0-hosting-service-account-list-resource`
         -

       * - ``info``
         - :ref:`webapi2.0-server-info-resource`
         -

       * - ``muted_review_request``
         - :ref:`webapi2.0-muted-review-request-resource`
         -

       * - ``muted_review_requests``
         - :ref:`webapi2.0-muted-review-request-list-resource`
         -

       * - ``oauth_app``
         - :ref:`webapi2.0-oauth-application-resource`
         -

       * - ``oauth_apps``
         - :ref:`webapi2.0-oauth-application-list-resource`
         -

       * - ``oauth_token``
         - :ref:`webapi2.0-oauth-token-resource`
         -

       * - ``oauth_tokens``
         - :ref:`webapi2.0-oauth-token-list-resource`
         -

       * - ``remote_repository``
         - :ref:`webapi2.0-remote-repository-resource`
         -

       * - ``remote_repositories``
         - :ref:`webapi2.0-remote-repository-list-resource`
         -

       * - ``repository``
         - :ref:`webapi2.0-repository-resource`
         -

       * - ``repositories``
         - :ref:`webapi2.0-repository-list-resource`
         -

       * - ``repository_branches``
         - :ref:`webapi2.0-repository-branches-resource`
         - 5.0.2

       * - ``repository_commits``
         - :ref:`webapi2.0-repository-commits-resource`
         - 5.0.2

       * - ``repository_group``
         - :ref:`webapi2.0-repository-group-resource`
         - 5.0.2

       * - ``repository_groups``
         - :ref:`webapi2.0-repository-group-list-resource`
         - 5.0.2

       * - ``repository_info``
         - :ref:`webapi2.0-repository-info-resource`
         - 5.0.2

       * - ``repository_user``
         - :ref:`webapi2.0-repository-user-resource`
         - 5.0.2

       * - ``repository_users``
         - :ref:`webapi2.0-repository-user-list-resource`
         - 5.0.2

       * - ``review``
         - :ref:`webapi2.0-review-resource`
         -

       * - ``reviews``
         - :ref:`webapi2.0-review-list-resource`
         -

       * - ``review_diff_comment``
         - :ref:`webapi2.0-review-diff-comment-resource`
         - 5.0.2

       * - ``review_diff_comments``
         - :ref:`webapi2.0-review-diff-comment-list-resource`
         - 5.0.2

       * - ``review_draft``
         - :ref:`webapi2.0-review-draft-resource`
         -

       * - ``review_general_comment``
         - :ref:`webapi2.0-review-general-comment-resource`
         - 5.0.2

       * - ``review_general_comments``
         - :ref:`webapi2.0-review-general-comment-list-resource`
         - 5.0.2

       * - ``review_group_user``
         - :ref:`webapi2.0-review-group-user-resource`
         -

       * - ``review_group_users``
         - :ref:`webapi2.0-review-group-user-list-resource`
         -

       * - ``review_reply``
         - :ref:`webapi2.0-review-reply-resource`
         - 5.0.2

       * - ``review_replies``
         - :ref:`webapi2.0-review-reply-list-resource`
         - 5.0.2

       * - ``review_reply_draft``
         - :ref:`webapi2.0-review-reply-draft-resource`
         - 5.0.2

       * - ``review_reply_diff_comment``
         - :ref:`webapi2.0-review-reply-diff-comment-resource`
         - 5.0.2

       * - ``review_reply_diff_comments``
         - :ref:`webapi2.0-review-reply-diff-comment-list-resource`
         - 5.0.2

       * - ``review_reply_file_attachment_comment``
         - :ref:`webapi2.0-review-reply-file-attachment-comment-resource`
         - 5.0.2

       * - ``review_reply_file_attachment_comments``
         - :ref:`webapi2.0-review-reply-file-attachment-comment-list-resource`
         - 5.0.2

       * - ``review_reply_general_comment``
         - :ref:`webapi2.0-review-reply-general-comment-resource`
         - 5.0.2

       * - ``review_reply_general_comments``
         - :ref:`webapi2.0-review-reply-general-comment-list-resource`
         - 5.0.2

       * - ``review_reply_screenshot_comment``
         - :ref:`webapi2.0-review-reply-screenshot-comment-resource`
         - 5.0.1

       * - ``review_reply_screenshot_comments``
         - :ref:`webapi2.0-review-reply-screenshot-comment-list-resource`
         - 5.0.1

       * - ``review_request``
         - :ref:`webapi2.0-review-request-resource`
         -

       * - ``review_requests``
         - :ref:`webapi2.0-review-request-list-resource`
         -

       * - ``review_request_change``
         - :ref:`webapi2.0-change-resource`
         - 5.0.2

       * - ``review_request_changes``
         - :ref:`webapi2.0-change-list-resource`
         - 5.0.2

       * - ``review_request_draft``
         - :ref:`webapi2.0-review-request-draft-resource`
         - 5.0.2

       * - ``review_request_file_attachment``
         - :ref:`webapi2.0-file-attachment-resource`
         - 5.0.2

       * - ``review_request_file_attachments``
         - :ref:`webapi2.0-file-attachment-list-resource`
         - 5.0.2

       * - ``review_request_file_attachment_comments``
         - :ref:`webapi2.0-file-attachment-comment-list-resource`
         - 5.0.2

       * - ``review_request_last_update``
         - :ref:`webapi2.0-review-request-last-update-resource`
         - 5.0.2

       * - ``review_request_status_update``
         - :ref:`webapi2.0-status-update-resource`
         - 5.0.2

       * - ``review_request_status_updates``
         - :ref:`webapi2.0-status-update-list-resource`
         - 5.0.2

       * - ``root``
         - :ref:`webapi2.0-root-resource`
         -

       * - ``screenshot``
         - :ref:`webapi2.0-screenshot-resource`
         -

       * - ``screenshots``
         - :ref:`webapi2.0-screenshot-list-resource`
         -

       * - ``screenshot_comment``
         - :ref:`webapi2.0-review-screenshot-comment-resource`
         -

       * - ``screenshot_comments``
         - :ref:`webapi2.0-screenshot-comment-list-resource`
         -

       * - ``search``
         - :ref:`webapi2.0-search-resource`
         -

       * - ``session``
         - :ref:`webapi2.0-session-resource`
         -

       * - ``user``
         - :ref:`webapi2.0-user-resource`
         -

       * - ``users``
         - :ref:`webapi2.0-user-list-resource`
         -

       * - ``user_file_attachment``
         - :ref:`webapi2.0-user-file-attachment-resource`
         -

       * - ``user_file_attachments``
         - :ref:`webapi2.0-user-file-attachment-list-resource`
         -

       * - ``validation``
         - :ref:`webapi2.0-validation-resource`
         -

       * - ``watched``
         - :ref:`webapi2.0-watched-resource`
         -

       * - ``watched_review_group``
         - :ref:`webapi2.0-watched-review-group-resource`
         -

       * - ``watched_review_groups``
         - :ref:`webapi2.0-watched-review-group-list-resource`
         -

       * - ``watched_review_request``
         - :ref:`webapi2.0-watched-review-request-resource`
         -

       * - ``watched_review_requests``
         - :ref:`webapi2.0-watched-review-request-list-resource`
         -

       * - ``webhook``
         - :ref:`webapi2.0-web-hook-resource`
         -

       * - ``webhooks``
         - :ref:`webapi2.0-web-hook-list-resource`
         -

    The following URI templates are considered deprecated, and may be
    removed in a future version.

    .. list-table::
       :header-rows: 1

       * - URI template key
         - Replacement
         - Resource
         - Deprecated In

       * - ``branches``
         - ``repository_branches``
         - :ref:`webapi2.0-repository-branches-resource`
         - 5.0.2

       * - ``change``
         - ``review_request_change``
         - :ref:`webapi2.0-change-resource`
         - 5.0.2

       * - ``changes``
         - ``review_request_changes``
         - :ref:`webapi2.0-change-list-resource`
         - 5.0.2

       * - ``commit``
         - ``diff_commit``
         - :ref:`webapi2.0-diff-commit-resource`
         - 5.0.2

       * - ``commits``
         - ``repository_commits``
         - :ref:`webapi2.0-repository-commits-resource`
         - 5.0.2

       * - ``diff_comment``
         - ``review_reply_diff_comment``
         - :ref:`webapi2.0-review-reply-diff-comment-resource`
         - 5.0.2

       * - ``diff_comments``
         - ``review_reply_diff_comments``
         - :ref:`webapi2.0-review-reply-diff-comment-list-resource`
         - 5.0.2

       * - ``draft``
         - ``review_request_draft``
         - :ref:`webapi2.0-review-request-draft-resource`
         - 5.0.2

       * - ``file``
         - ``file_diff``
         - :ref:`webapi2.0-file-diff-resource`
         - 5.0.2

       * - ``files``
         - ``file_diffs``
         - :ref:`webapi2.0-file-diff-list-resource`
         - 5.0.2

       * - ``file_attachment``
         - ``review_request_file_attachment``
         - :ref:`webapi2.0-file-attachment-resource`
         - 5.0.2

       * - ``file_attachments``
         - ``review_request_file_attachments``
         - :ref:`webapi2.0-file-attachment-list-resource`
         - 5.0.2

       * - ``general_comment``
         - ``review_general_comment``
         - :ref:`webapi2.0-review-general-comment-resource`
         - 5.0.2

       * - ``general_comments``
         - ``review_general_comments``
         - :ref:`webapi2.0-review-general-comment-list-resource`
         - 5.0.2

       * - ``last_update``
         - ``review_request_last_update``
         - :ref:`webapi2.0-review-request-last-update-resource`
         - 5.0.2

       * - ``original_file``
         - ``file_diff_original_file``
         - :ref:`webapi2.0-original-file-resource`
         - 5.0.2

       * - ``patched_file``
         - ``file_diff_patched_file``
         - :ref:`webapi2.0-patched-file-resource`
         - 5.0.2

       * - ``reply``
         - ``review_reply``
         - :ref:`webapi2.0-review-reply-resource`
         - 5.0.2

       * - ``replies``
         - ``review_replies``
         - :ref:`webapi2.0-review-reply-list-resource`
         - 5.0.2

       * - ``reply_draft``
         - ``review_reply_draft``
         - :ref:`webapi2.0-review-reply-draft-resource`
         - 5.0.2

       * - ``status_update``
         - ``review_request_status_update``
         - :ref:`webapi2.0-status-update-resource`
         - 5.0.2

       * - ``status_updates``
         - ``review_request_status_update``
         - :ref:`webapi2.0-status-update-list-resource`
         - 5.0.2
    """

    mimetype_vendor = 'reviewboard.org'

    def __init__(self, *args, **kwargs):
        super(RootResource, self).__init__([
            resources.default_reviewer,
            resources.extension,
            resources.hosting_service,
            resources.hosting_service_account,
            resources.oauth_app,
            resources.oauth_token,
            resources.repository,
            resources.review_group,
            resources.review_request,
            resources.root_diff_comment,
            resources.root_file_attachment_comment,
            resources.root_general_comment,
            resources.root_review,
            resources.search,
            resources.server_info,
            resources.session,
            resources.user,
            resources.validation,
            resources.webhook,
        ], *args, **kwargs)

        # Manually include these resources to maintain compatibility with
        # our Python 2.7 API behavior. This is a bandaid for a larger
        # issue that stems from resources that share the same name but
        # have different URI templates.
        self.register_uri_template(
            name='branches',
            relative_path='repositories/{repository_id}/branches/',
            relative_resource=self)
        self.register_uri_template(
            name='change',
            relative_path=(
                'review-requests/{review_request_id}/changes/{change_id}/'
            ),
            relative_resource=self)
        self.register_uri_template(
            name='changes',
            relative_path='review-requests/{review_request_id}/changes/',
            relative_resource=self)
        self.register_uri_template(
            name='commit',
            relative_path=(
                'review-requests/{review_request_id}/diffs/'
                '{diff_revision}/commits/{commit_id}/'
            ),
            relative_resource=self)
        self.register_uri_template(
            name='commits',
            relative_path='repositories/{repository_id}/commits/',
            relative_resource=self)
        self.register_uri_template(
            name='draft',
            relative_path='review-requests/{review_request_id}/draft/',
            relative_resource=self)
        self.register_uri_template(
            name='file',
            relative_path=(
                'review-requests/{review_request_id}/diffs/'
                '{diff_revision}/files/{filediff_id}/'
            ),
            relative_resource=self)
        self.register_uri_template(
            name='files',
            relative_path=(
                'review-requests/{review_request_id}/diffs/'
                '{diff_revision}/files/'
            ),
            relative_resource=self)
        self.register_uri_template(
            name='diff_comment',
            relative_path=(
                'review-requests/{review_request_id}/reviews/'
                '{review_id}/replies/{reply_id}/diff-comments/'
                '{comment_id}/'
            ),
            relative_resource=self)
        self.register_uri_template(
            name='diff_comments',
            relative_path=(
                'review-requests/{review_request_id}/reviews/'
                '{review_id}/replies/{reply_id}/diff-comments/'
            ),
            relative_resource=self)
        self.register_uri_template(
            name='file_attachment',
            relative_path=(
                'review-requests/{review_request_id}/file-attachments/'
                '{file_attachment_id}/'
            ),
            relative_resource=self)
        self.register_uri_template(
            name='file_attachments',
            relative_path=(
                'review-requests/{review_request_id}/file-attachments/'
            ),
            relative_resource=self)
        self.register_uri_template(
            name='general_comments',
            relative_path=(
                'review-requests/{review_request_id}/reviews/'
                '{review_id}/general-comments/'
            ),
            relative_resource=self)
        self.register_uri_template(
            name='general_comment',
            relative_path=(
                'review-requests/{review_request_id}/reviews/'
                '{review_id}/general-comments/{comment_id}/'
            ),
            relative_resource=self)
        self.register_uri_template(
            name='last_update',
            relative_path='review-requests/{review_request_id}/last-update/',
            relative_resource=self)
        self.register_uri_template(
            name='original_file',
            relative_path=(
                'review-requests/{review_request_id}/diffs/{diff_revision}/'
                'files/{filediff_id}/original-file/'
            ),
            relative_resource=self)
        self.register_uri_template(
            name='patched_file',
            relative_path=(
                'review-requests/{review_request_id}/diffs/{diff_revision}/'
                'files/{filediff_id}/patched-file/'
            ),
            relative_resource=self)
        self.register_uri_template(
            name='reply',
            relative_path=(
                'review-requests/{review_request_id}/reviews/{review_id}/'
                'replies/{reply_id}/'
            ),
            relative_resource=self)
        self.register_uri_template(
            name='replies',
            relative_path=(
                'review-requests/{review_request_id}/reviews/{review_id}/'
                'replies/'
            ),
            relative_resource=self)
        self.register_uri_template(
            name='reply_draft',
            relative_path=(
                'review-requests/{review_request_id}/reviews/{review_id}/'
                'replies/draft/'
            ),
            relative_resource=self)
        self.register_uri_template(
            name='search',
            relative_path='search/{username}/',
            relative_resource=self)
        self.register_uri_template(
            name='status_update',
            relative_path=(
                'review-requests/{review_request_id}/status-updates/'
                '{status_update_id}/'
            ),
            relative_resource=self)
        self.register_uri_template(
            name='status_updates',
            relative_path=(
                'review-requests/{review_request_id}/status-updates/'
            ),
            relative_resource=self)

    @webapi_check_login_required
    @webapi_check_local_site
    @augment_method_from(DjbletsRootResource)
    def get(self, request, *args, **kwargs):
        """Retrieves the list of top-level resources and templates."""
        pass

    def serialize_root(self, request, *args, **kwargs):
        root = super(RootResource, self).serialize_root(request, *args,
                                                        **kwargs)
        root.update(get_server_info(request))

        return root


root_resource = RootResource()



"""A hook for creating new thumbnailers for file attachments."""

from __future__ import annotations

from djblets.extensions.hooks import ExtensionHook, ExtensionHookPoint

from reviewboard.attachments.mimetypes import (register_mimetype_handler,
                                               unregister_mimetype_handler)


class FileAttachmentThumbnailHook(ExtensionHook, metaclass=ExtensionHookPoint):
    """This hook allows custom thumbnails to be defined for file attachments.

    This accepts a list of mimetype handlers specified by the Extension
    that must:

    * Subclass :py:class:`reviewboard.attachments.mimetypes.MimetypeHandler`
    * Define a list of file mimetypes it can handle in a class variable
      called ``supported_mimetypes``
    * Define how to generate a thumbnail of that mimetype by overriding
      the instance function ``def get_thumbnail(self):``

    These mimetype handlers are registered when the hook is created. Likewise,
    it unregisters the same list of mimetype handlers when the extension is
    disabled.
    """

    def initialize(self, mimetype_handlers):
        """Initialize the hook.

        This will register each of the provided mimetype handler classes.

        Args:
            mimetype_handlers (list of type):
                The list of mimetype handlers to register. Each must be a
                subclass of
                :py:class:`~reviewboard.attachments.mimetypes.MimetypeHandler`.

        Raises:
            TypeError:
                One or more of the provided classes are not of the correct
                type.
        """
        self.mimetype_handlers = mimetype_handlers

        for mimetype_handler in self.mimetype_handlers:
            register_mimetype_handler(mimetype_handler)

    def shutdown(self):
        """Shut down the hook.

        This will unregister each of the mimetype handler classes.
        """
        for mimetype_handler in self.mimetype_handlers:
            unregister_mimetype_handler(mimetype_handler)
